"""Fixer that inserts mypy annotations from json file into code.

This fixer consumes json from TYPE_COLLECTION_JSON env variable in the following format:

[
    {
        "path": "/Users/svorobev/src/client/build_number/__init__.py",
        "func_name": "is_test",
        "arg_types": ["int", "str"],
        "ret_type": "Any"
    },
    ...
]

(The old format with "type_comment" instead of "arg_types" and
"ret_type" is also still supported.)
"""

from __future__ import print_function

import json  # noqa
import os
from lib2to3.pytree import Base, Leaf, Node
from typing import Any, Dict, List, Optional, Tuple, Union

try:
    from typing import Text
except ImportError:
    # In Python 3.5.1 stdlib, typing.py does not define Text
    Text = str  # type: ignore

from .base import BaseFixAnnotateFromSignature


class FixAnnotateJson(BaseFixAnnotateFromSignature):

    def get_types(self, node, results, funcname):
        # type: (Union[Leaf, Node], Dict[str, Any], str) -> Optional[Tuple[List[str], str]]
        data = self.type_options['type_info']
        top_dir = self.type_options['top_dir']
        # FIXME: this lookup code is inefficient and fragile.  We should restructure
        #  the json to be keyed by (funcname, rel path), and compute a rel path from self.filename.
        # We are using relative paths in the JSON.
        items = [
            it for it in data
            if it['func_name'] == funcname and
            (it['path'] == self.filename or
                os.path.join(top_dir, it['path']) == os.path.abspath(self.filename))
        ]
        if len(items) > 1:
            # this can happen, because of
            # 1) nested functions
            # 2) method decorators
            # as a cheap and dirty solution we just return the nearest one by the line number
            # (keep the commented-out log_message call in case we need to come back to this)
            # self.log_message("%s:%d: duplicate signatures for %s (at lines %s)" %
            # (items[0]['path'], node.get_lineno(), items[0]['func_name'],
            # ", ".join(str(it['line']) for it in items)))
            items.sort(key=lambda it: abs(node.get_lineno() - it['line']))
        if items:
            it = items[0]
            # If the line number is too far off, the source probably drifted
            # since the trace was collected; it's better to skip this node.
            # (Allow some drift, since decorators also cause an offset.)
            if abs(node.get_lineno() - it['line']) >= self.line_drift:
                self.log_message("%s:%d: '%s' signature from line %d too far away -- skipping" %
                                 (self.filename, node.get_lineno(), it['func_name'], it['line']))
                return None
            if 'signature' in it:
                return it['signature']['arg_types'], it['signature']['return_type']
        return None



from __future__ import absolute_import, print_function

from lib2to3 import pytree
from lib2to3.fixer_util import does_tree_import, syms
from lib2to3.pgen2 import token
from lib2to3.pytree import Node
from typing import Any, Dict, List, Match, Optional, Tuple, cast

from ..docs import formats
from .base import get_funcname, typing_all
from .fix_annotate_json import BaseFixAnnotateFromSignature

SPECIAL_METHOD_RETURN = {
    '__init__': 'None'
}


def _get_type(typ, default='Any'):
    # type: (Optional[str], str) -> str
    return default if typ is None else typ


def find_classdef(node):
    # type: (pytree.Node) -> Optional[List[pytree.Node]]
    """
    Parameters
    ----------
    node : pytree.Node

    Returns
    -------
    Optional[List[pytree.Node]]
    """
    while True:
        parent = node.parent
        if parent is None:
            return None
        elif parent.type == syms.classdef:
            # return the suite as list
            for child in parent.children:
                if child.type == syms.suite:
                    return [cast(Node, child)]
            else:
                raise RuntimeError("could not find suite")
        node = parent


def get_docstring(suite):
    # type: (list) -> Tuple[Optional[str], int]
    """
    Get docstring data from a lib2t3 suite object

    Parameters
    ----------
    suite : list

    Returns
    -------
    docstring : Optional[str]
    lineno : int
    """
    assert isinstance(suite, list)
    if suite[0].children[1].type == token.INDENT:
        indent_node = suite[0].children[1]
        doc_node = suite[0].children[2]
    else:
        # e.g. "def foo(...): x = 5; y = 7"
        return None, -1

    if isinstance(doc_node, pytree.Node) and \
            doc_node.children[0].type == token.STRING:
        leaf = doc_node.children[0]
        if not isinstance(leaf, pytree.Leaf):
            raise RuntimeError
        doc = leaf.value
        # convert '"docstring"' to 'docstring'
        # FIXME: something better than eval
        return eval(doc), leaf.lineno
    else:
        # no docstring
        return None, -1


def keep_arg(i, arg_name, typ):
    # type: (int, str, Optional[str]) -> bool
    """
    Return whether `arg_name` should be included in the type annotation.

    Parameters
    ----------
    i : int
        position within the arg list
    arg_name : str
    typ : Optional[str]
        type of the arg

    Returns
    -------
    bool
    """
    # we can't *easily* tell from here if the current func is a
    # method, but the below is a pretty good assurance that we can skip
    # the current arg.
    # there is nothing wrong with including self or cls, but pep484
    # supports omitting it for brevity.
    return not (i == 0 and arg_name in {'self', 'cls'} and typ is None)


class FixAnnotateDocs(BaseFixAnnotateFromSignature):
    """Inserts annotations by parsing docstrings."""

    def get_format(self):
        # type: () -> str
        """
        Returns
        -------
        str
        """
        return self.type_options['doc_format']

    def get_default_return_type(self):
        # type: () -> str
        """
        Returns
        -------
        str
        """
        return self.type_options['doc_default_return_type']

    def parse_docstring(self, docstring, line):
        # type: (str, int) -> Tuple[Dict[str, str], Optional[str]]
        """
        Parameters
        ----------
        docstring : str
        line : int

        Returns
        -------
        args : Dict[str, str]
        result : Optional[str]
        """
        if docstring:
            params, result = formats.parse_docstring(docstring, line=line,
                                                     format_name=self.get_format(),
                                                     filename=self.filename)
            return ({k: v.type for k, v in params.items() if v.type},
                    result.type if result else None)
        else:
            return {}, None

    def should_skip(self, node, results):
        return False

    def type_updater(self, match, node):
        # type: (Match, Node) -> str
        # Replace `pkg.mod.SomeClass` with `SomeClass`
        # and remember to import it.
        word = match.group()
        if does_tree_import(None, word, node):
            # Check whether there already exists an import binding for this
            return word
        # If not, assume it's either builtin or from `typing`
        self.touch_typing_import(word, node)
        return word

    def make_annotation(self, node, results):
        # type: (Node, Dict[str, Any]) -> Optional[Tuple[List[str], str]]
        suite = results["suite"]

        docstring, line = get_docstring(suite)

        args = results.get("args")
        name_node = results["name"]

        funcname = get_funcname(node)
        short_funcname = funcname.rsplit('.')[-1]

        class_suite = find_classdef(name_node)
        is_method = class_suite is not None

        if docstring is None and short_funcname == '__init__' and class_suite is not None:
            # fall back to the class docstring
            docstring, line = get_docstring(class_suite)

        if docstring is None:
            return None

        arg_types = []  # type: List[str]
        argtype_map, ret_type = self.parse_docstring(docstring, line)

        if not argtype_map and not ret_type:
            # if the user has provided type annoations, we always use them
            return None

        if is_method and short_funcname in SPECIAL_METHOD_RETURN:
            default_return = SPECIAL_METHOD_RETURN[short_funcname]
        else:
            default_return = self.get_default_return_type()

        ret_type = _get_type(ret_type, default_return)

        if args:
            # if args.type == syms.tfpdef:
            #     pass
            if args.type == syms.typedargslist:
                arg_list = []  # type: List[str]
                kind_list = []  # type: List[str]
                consume = True
                kind = ''
                for arg in args.children:
                    if consume and arg.type == token.NAME:
                        arg_list.append(arg.value)
                        kind_list.append(kind)
                        consume = False
                    elif consume and arg.type == token.STAR:
                        kind = '*'
                    elif consume and arg.type == token.DOUBLESTAR:
                        kind = '**'
                    elif arg.type == token.COMMA:
                        consume = True
                        kind = ''
            elif args.type == token.NAME:
                arg_list = [args.value]
                kind_list = ['']
            else:
                raise TypeError(args)

            for i, (arg_name, kind) in enumerate(zip(arg_list, kind_list)):
                typ = argtype_map.get(arg_name)
                if typ is None and formats.default_arg_types:
                    typ = formats.default_arg_types.get(arg_name)
                if not is_method or keep_arg(i, arg_name, typ):
                    arg_types.append(kind + _get_type(typ).strip('*'))

        arg_types = [self.update_type_names(arg_type, node) for arg_type in arg_types]
        ret_type = self.update_type_names(ret_type, node)
        return arg_types, ret_type



"""Fixer that inserts mypy annotations into all methods.

This transforms e.g.

  def foo(self, bar, baz=12):
      return bar + baz

into a type annoted version:

	  def foo(self, bar, baz=12):
	      # type: (Any, int) -> Any            # noqa: F821
	      return bar + baz

or (when setting options['typeright']['annotation_style'] to 'py3'):

	  def foo(self, bar : Any, baz : int = 12) -> Any:
	      return bar + baz


It does not do type inference but it recognizes some basic default
argument values such as numbers and strings (and assumes their type
implies the argument type).

It also uses some basic heuristics to decide whether to ignore the
first argument:

  - always if it's named 'self'
  - if there's a @classmethod decorator

Finally, it knows that __init__() is supposed to return None.
"""

from __future__ import print_function

import os
import re
from contextlib import contextmanager
from lib2to3.fixer_base import BaseFix
from lib2to3.fixer_util import (does_tree_import, find_indentation, syms,
                                touch_import)
from lib2to3.patcomp import compile_pattern
from lib2to3.pgen2 import token
from lib2to3.pytree import Base, Leaf, Node
from typing import (Any, Dict, Iterator, List, Match, Optional, Set, Text,
                    Tuple, Union)
from typing import __all__ as typing_all  # type: ignore

from typeright.fixes.fixer_utils import (create_import,
                                          create_type_checking_import,
                                          get_unprotected_imports,
                                          type_by_import_stmt)

# Taken from mypy codebase:
# https://github.com/python/mypy/blob/745d300b8304c3dcf601477762bf9d70b9a4619c/mypy/main.py#L503

PY_EXTENSIONS = ['.pyi', '.py']
TYPE_REG = re.compile('\s*#\s*type:.*')


def crawl_up(arg):
    # type: (str) -> Tuple[str, str]
    """Given a .py[i] filename, return (root directory, module).
    We crawl up the path until we find a directory without
    __init__.py[i], or until we run out of path components.
    """
    dir, mod = os.path.split(arg)
    mod = strip_py(mod) or mod
    while dir and get_init_file(dir):
        dir, base = os.path.split(dir)
        if not base:
            break
        if mod == '__init__' or not mod:
            mod = base
        else:
            mod = base + '.' + mod
    return dir, mod


def strip_py(arg):
    # type: (str) -> Optional[str]
    """Strip a trailing .py or .pyi suffix.
    Return None if no such suffix is found.
    """
    for ext in PY_EXTENSIONS:
        if arg.endswith(ext):
            return arg[:-len(ext)]
    return None


def get_init_file(dir):
    # type: (str) -> Optional[str]
    """Check whether a directory contains a file named __init__.py[i].
    If so, return the file's name (with dir prefixed).  If not, return
    None.
    This prefers .pyi over .py (because of the ordering of PY_EXTENSIONS).
    """
    for ext in PY_EXTENSIONS:
        f = os.path.join(dir, '__init__' + ext)
        if os.path.isfile(f):
            return f
    return None


def get_funcname(node):
    # type: (Optional[Union[Leaf, Node]]) -> Text
    """Get function name by (approximately) the following rules:

    - function -> function_name
    - method -> ClassName.function_name

    More specifically, we include every class and function name that
    the node is a child of, so nested classes and functions get names like
    OuterClass.InnerClass.outer_fn.inner_fn.
    """
    components = []  # type: List[str]
    while node:
        if node.type in (syms.classdef, syms.funcdef):
            name = node.children[1]
            assert name.type == token.NAME, repr(name)
            assert isinstance(name, Leaf)  # Same as previous, for mypy
            components.append(name.value)
        node = node.parent
    return '.'.join(reversed(components))


def count_args(node, results):
    # type: (Node, Dict[str, Base]) -> Tuple[int, bool, bool, bool]
    """Count arguments and check for self and *args, **kwds.

    Return (selfish, count, star, starstar) where:
    - count is total number of args (including *args, **kwds)
    - selfish is True if the initial arg is named 'self' or 'cls'
    - star is True iff *args is found
    - starstar is True iff **kwds is found
    """
    count = 0
    selfish = False
    star = False
    starstar = False
    args = results.get('args')
    if isinstance(args, Node):
        children = args.children
    elif isinstance(args, Leaf):
        children = [args]
    else:
        children = []
    # Interpret children according to the following grammar:
    # (('*'|'**')? NAME ['=' expr] ','?)*
    skip = False
    previous_token_is_star = False
    for child in children:
        if skip:
            skip = False
        elif isinstance(child, Leaf):
            # A single '*' indicates the rest of the arguments are keyword only
            # and shouldn't be counted as a `*`.
            if child.type == token.STAR:
                previous_token_is_star = True
            elif child.type == token.DOUBLESTAR:
                starstar = True
            elif child.type == token.NAME:
                if count == 0:
                    if child.value in ('self', 'cls'):
                        selfish = True
                count += 1
                if previous_token_is_star:
                    star = True
            elif child.type == token.EQUAL:
                skip = True
            if child.type != token.STAR:
                previous_token_is_star = False
    return count, selfish, star, starstar


def is_type_comment(comment):
    return TYPE_REG.match(comment)


class BaseFixAnnotate(BaseFix):

    # This fixer is compatible with the bottom matcher.
    BM_compatible = True

    # This fixer shouldn't run by default.
    explicit = True

    # The pattern to match.
    PATTERN = """
              funcdef< 'def' name=any parameters=parameters< '(' [args=any] rpar=')' > ':' suite=any+ >
              """

    _maxfixes = os.getenv('MAXFIXES')
    counter = None if not _maxfixes else int(_maxfixes)
    _type_options = None  # type: Optional[Dict[str, Any]]

    @property
    def type_options(self):
        if self._type_options is None:
            self._type_options = self.options.get('typeright', {})
        return self._type_options

    def should_skip(self, node, results):
        if BaseFixAnnotate.counter is not None:
            if BaseFixAnnotate.counter <= 0:
                return True

        # Check if there's already a long-form annotation for some argument.
        parameters = results.get('parameters')
        if parameters is not None:
            for ch in parameters.pre_order():
                if ch.prefix.lstrip().startswith('# type:'):
                    return True

        args = results.get('args')
        if args is not None:
            for ch in args.pre_order():
                if ch.prefix.lstrip().startswith('# type:'):
                    return True

        children = results['suite'][0].children

        # NOTE: I've reverse-engineered the structure of the parse tree.
        # It's always a list of nodes, the first of which contains the
        # entire suite.  Its children seem to be:
        #
        #   [0] NEWLINE
        #   [1] INDENT
        #   [2...n-2] statements (the first may be a docstring)
        #   [n-1] DEDENT
        #
        # Comments before the suite are part of the INDENT's prefix.
        #
        # "Compact" functions (e.g. "def foo(x, y): return max(x, y)")
        # have a different structure (no NEWLINE, INDENT, or DEDENT).

        # Check if there's already an annotation.
        for ch in children:
            if ch.prefix.lstrip().startswith('# type:'):
                return True  # There's already a # type: comment here; don't change anything.

        # Python 3 style return annotation are already skipped by the pattern

        # Python 3 style argument annotation structure
        #
        # Structure of the arguments tokens for one positional argument without default value :
        # + LPAR '('
        # + NAME_NODE_OR_LEAF arg1
        # + RPAR ')'
        #
        # NAME_NODE_OR_LEAF is either:
        # 1. Just a leaf with value NAME
        # 2. A node with children: NAME, ':", node expr or value leaf
        #
        # Structure of the arguments tokens for one args with default value or multiple
        # args, with or without default value, and/or with extra arguments :
        # + LPAR '('
        # + node
        #   [
        #     + NAME_NODE_OR_LEAF
        #      [
        #        + EQUAL '='
        #        + node expr or value leaf
        #      ]
        #    (
        #        + COMMA ','
        #        + NAME_NODE_OR_LEAF positional argn
        #      [
        #        + EQUAL '='
        #        + node expr or value leaf
        #      ]
        #    )*
        #   ]
        #   [
        #     + STAR '*'
        #     [
        #     + NAME_NODE_OR_LEAF positional star argument name
        #     ]
        #   ]
        #   [
        #     + COMMA ','
        #     + DOUBLESTAR '**'
        #     + NAME_NODE_OR_LEAF positional keyword argument name
        #   ]
        # + RPAR ')'

        # Let's skip Python 3 argument annotations
        it = iter(args.children) if args else iter([])
        for ch in it:
            if ch.type == token.STAR:
                # *arg part
                ch = next(it)
                if ch.type == token.COMMA:
                    continue
            elif ch.type == token.DOUBLESTAR:
                # *arg part
                ch = next(it)
            if ch.type > 256:
                # this is a node, therefore an annotation
                assert ch.children[0].type == token.NAME
                return True
            try:
                ch = next(it)
                if ch.type == token.COLON:
                    # this is an annotation
                    return True
                elif ch.type == token.EQUAL:
                    ch = next(it)
                    ch = next(it)
                assert ch.type == token.COMMA
                continue
            except StopIteration:
                break

        return False

    def transform(self, node, results):
        if self.should_skip(node, results):
            return

        # Compute the annotation
        annot = self.make_annotation(node, results)
        if annot is None:
            return
        argtypes, restype = annot

        if self.type_options['annotation_style'] == 'py3':
            self.add_py3_annot(argtypes, restype, node, results)
        else:
            self.add_py2_annot(argtypes, restype, node, results)

        # Common to py2 and py3 style annotations:
        if BaseFixAnnotate.counter is not None:
            BaseFixAnnotate.counter -= 1

        # Also add 'from typing import Any' at the top if needed.
        self.patch_imports(argtypes + [restype], node)

    def add_py3_annot(self, argtypes, restype, node, results):
        # type: (List[str], str, Node, Dict[str, Any]) -> None

        args = results.get('args')  # type: Optional[Node]

        argleaves = []  # type: List[Tuple[str, Leaf]]
        if args is None:
            # function with 0 arguments
            it = iter([])  # type: Iterator[Union[Leaf, Node]]
        elif len(args.children) == 0:
            # function with 1 argument
            it = iter([args])
        else:
            # function with multiple arguments or 1 arg with default value
            it = iter(args.children)

        for ch in it:
            argstyle = 'name'
            if ch.type == token.STAR:
                # *arg part
                argstyle = 'star'
                ch = next(it)
                if ch.type == token.COMMA:
                    continue
            elif ch.type == token.DOUBLESTAR:
                # *arg part
                argstyle = 'keyword'
                ch = next(it)
            assert ch.type == token.NAME
            assert isinstance(ch, Leaf)
            argleaves.append((argstyle, ch))
            try:
                ch = next(it)
                if ch.type == token.EQUAL:
                    ch = next(it)
                    ch = next(it)
                assert ch.type == token.COMMA
                continue
            except StopIteration:
                break

        # when self or cls is not annotated, argleaves == argtypes+1
        argleaves = argleaves[len(argleaves) - len(argtypes):]

        for ch_withstyle, chtype in zip(argleaves, argtypes):
            style, ch = ch_withstyle
            if style == 'star':
                assert chtype[0] == '*'
                assert chtype[1] != '*'
                chtype = chtype[1:]
            elif style == 'keyword':
                assert chtype[0:2] == '**'
                assert chtype[2] != '*'
                chtype = chtype[2:]
            ch.value = '%s: %s' % (ch.value, chtype)

            # put spaces around the equal sign
            if ch.next_sibling and ch.next_sibling.type == token.EQUAL:
                nextch = ch.next_sibling
                if not nextch.prefix[:1].isspace():
                    nextch.prefix = ' ' + nextch.prefix
                nextch_ = nextch.next_sibling
                assert nextch_ is not None
                if not nextch_.prefix[:1].isspace():
                    nextch_.prefix = ' ' + nextch_.prefix

        # Add return annotation
        rpar = results['rpar']
        rpar.value = '%s -> %s' % (rpar.value, restype)

        rpar.changed()

    def use_py2_long_form(self, argtypes, short_str, degen_str):
        # type: (List[str], str, str) -> bool
        if self.type_options['comment_style'] == 'single':
            return False
        elif self.type_options['comment_style'] == 'multi':
            return False
        else:  # auto
            return ((len(short_str) > 64 or len(argtypes) > 5)
                    and len(short_str) > len(degen_str))

    def add_py2_annot(self, argtypes, restype, node, results):
        # type: (List[str], str, Node, Dict[str, Any]) -> None

        children = results['suite'][0].children

        # Insert '# type: {annot}' comment.
        # For reference, see lib2to3/fixes/fix_tuple_params.py in stdlib.
        if len(children) >= 1 and children[0].type != token.NEWLINE:
            # one liner function
            if children[0].prefix.strip() == '':
                children[0].prefix = ''
                children.insert(0, Leaf(token.NEWLINE, '\n'))
                children.insert(
                    1, Leaf(token.INDENT, find_indentation(node) + '    '))
                children.append(Leaf(token.DEDENT, ''))

        if len(children) >= 2 and children[1].type == token.INDENT:
            degen_str = '(...) -> %s' % restype
            short_str = '(%s) -> %s' % (', '.join(argtypes), restype)
            if self.use_py2_long_form(argtypes, short_str, degen_str):
                self.insert_long_form(node, results, argtypes)
                annot_str = degen_str
            else:
                annot_str = short_str

            indent_node = children[1]
            comment, sep, other_comments = indent_node.prefix.partition('\n')
            comment = comment.rstrip() + sep
            annot_str = '# type: %s\n' % (annot_str,)
            if comment == annot_str:
                return

            if comment and not is_type_comment(comment):
                # push existing non-type comment to next line
                annot_str += comment

            indent_node.prefix = indent_node.value + annot_str + other_comments
            indent_node.changed()
        else:
            self.log_message("%s:%d: cannot insert annotation for one-line function" %
                             (self.filename, node.get_lineno()))

    def insert_long_form(self, node, results, argtypes):
        # type: (Node, Dict[str, Any], List[str]) -> None

        argtypes = list(argtypes)  # We destroy it
        args = results['args']
        if isinstance(args, Node):
            children = args.children
        elif isinstance(args, Leaf):
            children = [args]
        else:
            children = []
        # Interpret children according to the following grammar:
        # (('*'|'**')? NAME ['=' expr] ','?)*
        flag = False  # Set when the next leaf should get a type prefix
        indent = ''  # Will be set by the first child

        def set_prefix(child):
            if argtypes:
                arg = argtypes.pop(0).lstrip('*')
            else:
                arg = 'Any'  # Somehow there aren't enough args
            if not arg:
                # Skip self (look for 'check_self' below)
                prefix = child.prefix.rstrip()
            else:
                prefix = '  # type: ' + arg
                old_prefix = child.prefix.strip()
                if old_prefix:
                    assert old_prefix.startswith('#')
                    prefix += '  ' + old_prefix
            child.prefix = prefix + '\n' + indent

        check_self = self.is_method(node)
        for child in children:
            if isinstance(child, Leaf):
                if check_self and child.type == token.NAME:
                    check_self = False
                    if child.value in ('self', 'cls'):
                        argtypes.insert(0, '')
                if not indent:
                    indent = ' ' * child.column
                if child.value == ',':
                    flag = True
                elif flag:
                    set_prefix(child)
                    flag = False

        need_comma = len(children) >= 1 and children[-1].type != token.COMMA
        if need_comma and len(children) >= 2:
            if (children[-1].type == token.NAME and
                    (children[-2].type in (token.STAR, token.DOUBLESTAR))):
                need_comma = False
        if need_comma:
            children.append(Leaf(token.COMMA, u","))
        # Find the ')' and insert a prefix before it too.
        parameters = args.parent
        close_paren = parameters.children[-1]
        assert close_paren.type == token.RPAR, close_paren
        set_prefix(close_paren)
        assert not argtypes, argtypes

    def patch_imports(self, types, node):
        for typ in types:
            if 'Any' in typ:
                touch_import('typing', 'Any', node)
                break

    def make_annotation(self, node, results):
        # type: (Node, Dict[str, Any]) -> Optional[Tuple[List[str], str]]
        """Return the type annotations.

        Given the current Note and the dictionary parsed from PATTERN
        return the annoations for the arguments and return types as strings
        """
        raise NotImplementedError

    # The parse tree has a different shape when there is a single
    # decorator vs. when there are multiple decorators.
    DECORATED = "decorated< (d=decorator | decorators< dd=decorator+ >) funcdef >"
    decorated = compile_pattern(DECORATED)

    def get_decorators(self, node):
        """Return a list of decorators found on a function definition.

        This is a list of strings; only simple decorators
        (e.g. @staticmethod) are returned.

        If the function is undecorated or only non-simple decorators
        are found, return [].
        """
        if node.parent is None:
            return []
        results = {}
        if not self.decorated.match(node.parent, results):
            return []
        decorators = results.get('dd') or [results['d']]
        decs = []
        for d in decorators:
            for child in d.children:
                if isinstance(child, Leaf) and child.type == token.NAME:
                    decs.append(child.value)
        return decs

    def is_method(self, node):
        """Return whether the node occurs (directly) inside a class."""
        node = node.parent
        while node is not None:
            if node.type == syms.classdef:
                return True
            if node.type == syms.funcdef:
                return False
            node = node.parent
        return False

    RETURN_EXPR = "return_stmt< 'return' any >"
    return_expr = compile_pattern(RETURN_EXPR)

    def has_return_exprs(self, node):
        """Traverse the tree below node looking for 'return expr'.

        Return True if at least 'return expr' is found, False if not.
        (If both 'return' and 'return expr' are found, return True.)
        """
        results = {}
        if self.return_expr.match(node, results):
            return True
        for child in node.children:
            if child.type not in (syms.funcdef, syms.classdef):
                if self.has_return_exprs(child):
                    return True
        return False

    YIELD_EXPR = "yield_expr< 'yield' [any] >"
    yield_expr = compile_pattern(YIELD_EXPR)

    def is_generator(self, node):
        """Traverse the tree below node looking for 'yield [expr]'."""
        results = {}
        if self.yield_expr.match(node, results):
            return True
        for child in node.children:
            if child.type not in (syms.funcdef, syms.classdef):
                if self.is_generator(child):
                    return True
        return False


class BaseFixAnnotateFromSignature(BaseFixAnnotate):

    line_drift = 5
    safe_imports = set(['typing'])

    def __init__(self, options, log):
        super(BaseFixAnnotateFromSignature, self).__init__(options, log)

        self.needed_imports = set()  # type: Set[Tuple[str, str]]
        self.needed_type_checking_imports = set()  # type: Set[Tuple[str, str]]

    @classmethod
    @contextmanager
    def max_line_drift_set(cls, max_drift):
        old_drift = cls.line_drift
        cls.line_drift = max_drift
        try:
            yield
        finally:
            cls.line_drift = old_drift

    def add_import(self, mod, name, in_type_checking=False):
        """
        Adds potential import according to the 'in_type_checking' flag. If true, and if the potential
        import is not in safe_imports, then we assume this potential import is liable to cause an
        import cycle and should therefore be added under a TYPE_CHECKING block. If false, we just
        assume that isn't a concern and we add the import to the top level of the file.
        """
        if mod == self.current_module():
            return

        if in_type_checking and mod not in self.safe_imports:
            # Pass to the type checking queue
            self.needed_type_checking_imports.add((mod, name))
            return

        self.needed_imports.add((mod, name))

    def touch_typing_import(self, word, node):
        # type: (str, Node) -> None
        """
        Checks whether there exists import for type <word> and if not adds it to the instance
        import list

        Parameters
        -----------
        word : str
        node : Node
        """
        if word in typing_all:
            if not type_by_import_stmt('typing', word, node):
                # No import statement was found, should import
                self.add_import('typing', word)

    def patch_imports(self, types, node):
        if self.needed_imports:
            for mod, name in sorted(self.needed_imports):
                create_import(mod, name, node)
        if self.needed_type_checking_imports:
            for mod, name in sorted(self.needed_type_checking_imports):
                create_type_checking_import(mod, name, node)
        self.needed_imports.clear()
        self.needed_type_checking_imports.clear()

    def set_filename(self, filename):
        super(BaseFixAnnotateFromSignature, self).set_filename(filename)
        self._current_module = crawl_up(filename)[1]

    def current_module(self):
        # type: () -> str
        """Return the dotted path of the module currently being transformed"""
        return self._current_module

    def get_types(self, node, results, funcname):
        # type: (Node, Dict[str, Any], str) -> Optional[Tuple[List[str], str]]
        """Return types for `funcname`."""
        raise NotImplementedError

    def make_annotation(self, node, results):
        # type: (Node, Dict[str, Any]) -> Optional[Tuple[List[str], str]]
        name = results['name']
        assert isinstance(name, Leaf), repr(name)
        assert name.type == token.NAME, repr(name)
        funcname = get_funcname(node)

        def make(node, results, funcname):
            # type: (Node, Any, str) -> Optional[Tuple[List[str], str]]
            sig_data = self.get_types(node, results, funcname)
            if sig_data:
                arg_types, ret_type = sig_data
                return self.process_types(node, results, arg_types, ret_type)
            return None

        res = make(node, results, funcname)
        # If we couldn't find an annotation and this is a classmethod or
        # staticmethod, try again with just the funcname, since the
        # type collector can't figure out class names for those.
        # (We try with the full name above first so that tools that *can* figure
        # that out, like dmypy suggest, can use it.)
        if not res:
            decs = self.get_decorators(node)
            if 'staticmethod' in decs or 'classmethod' in decs:
                res = make(node, results, name.value)
        return res

    def process_types(self, node, results, arg_types, ret_type):
        # type: (Node, Dict[str, Any], List[str], str) -> Optional[Tuple[List[str], str]]
        """Process type annotations to handle star args, self-type, and imports."""
        # Passes 1-2 don't always understand *args or **kwds,
        # so add '*Any' or '**Any' at the end if needed.
        count, selfish, star, starstar = count_args(node, results)
        for arg_type in arg_types:
            if arg_type.startswith('**'):
                starstar = False
            elif arg_type.startswith('*'):
                star = False
        if star:
            arg_types.append('*Any')
        if starstar:
            arg_types.append('**Any')
        # Pass 1 omits the first arg iff it's named 'self' or 'cls',
        # even if it's not a method, so insert `Any` as needed
        # (but only if it's not actually a method).
        if selfish and len(arg_types) == count - 1:
            if self.is_method(node):
                count -= 1  # Leave out the type for 'self' or 'cls'
            else:
                arg_types.insert(0, 'Any')
        # If after those adjustments the count is still off,
        # print a warning and skip this node.
        if len(arg_types) != count:
            self.log_message("%s:%d: source has %d args, annotation has %d -- skipping" %
                             (self.filename, node.get_lineno(), count, len(arg_types)))
            return None

        arg_types = [self.update_type_names(arg_type, node) for arg_type in arg_types]
        # Avoid common error "No return value expected"
        if ret_type == 'None' and self.has_return_exprs(node):
            ret_type = 'Optional[Any]'
        # Special case for generators.
        if (self.is_generator(node) and
                not (ret_type == 'Iterator' or ret_type.startswith('Iterator['))):
            if ret_type.startswith('Optional['):
                assert ret_type[-1] == ']'
                ret_type = ret_type[9:-1]
            ret_type = 'Iterator[%s]' % ret_type
        ret_type = self.update_type_names(ret_type, node)
        return arg_types, ret_type

    def update_type_names(self, type_str, node):
        # type: (str, Node) -> str
        """Fixup module names and add necessary imports."""
        # Replace e.g. `List[pkg.mod.SomeClass]` with
        # `List[SomeClass]` and remember to import it.
        return re.sub(r'[\w.:]+', lambda m: self.type_updater(m, node), type_str)

    def type_updater(self, match, node):
        # type: (Match, Node) -> str
        # If import does not exist yet, replace `pkg.mod.SomeClass` with `SomeClass`
        # and remember to import it. Otherwise replace `pkg.mod.SomeClass` with the name
        # bound to its import
        word = match.group()
        if word == '...':
            return word
        if '.' not in word and ':' not in word:
            # Assume it's either builtin or from `typing`
            self.touch_typing_import(word, node)
            return word
        # If there is a :, treat that as the separator between the
        # module and the class.  Otherwise assume everything but the
        # last element is the module.
        if ':' in word:
            mod, name = word.split(':')
            to_import = name.split('.', 1)[0]
        else:
            mod, name = word.rsplit('.', 1)
            to_import = name

        # Get the typename we need to use to be valid for a current import statement
        result = type_by_import_stmt(mod, to_import, node)

        if result is not None:
            # There exists a current import statement this type is valid for
            # We know this module has already been imported
            return result

        in_type_checking = True
        # Get a list of cached imports from this node's root
        unprotected_imports = get_unprotected_imports(node)
        if mod in unprotected_imports:
            # If this mod was already imported in the original file, its safe to import from again
            in_type_checking = False

        # There exists no import statement for this type, add one
        self.add_import(mod, to_import, in_type_checking=in_type_checking)

        return name



# flake8: noqa
# Our flake extension misfires on type comments in strings below.

import json
import subprocess

from typeright.fixes.fix_annotate_command import FixAnnotateCommand
from typeright.fixes.tests import base_py2

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


class TestFixAnnotateCommand(base_py2.AnnotateFromSignatureTestCase):

    def setUp(self):
        super(TestFixAnnotateCommand, self).setUp(
            fix_list=["annotate_command"],
            fixer_pkg="typeright",
            options={
                'typeright': {
                    'annotation_style': 'py2',
                    'comment_style': 'auto',
                    'command': "fake {funcname} {filename}",
                },
            },
        )
        self.patcher = None

    def tearDown(self):
        if self.patcher is not None:
            self.patcher.stop()
            self.patcher = None
        super(TestFixAnnotateCommand, self).tearDown()

    def setTestData(self, data):
        self.filename = data[0]["path"]

        by_func = {d['func_name']: json.dumps([d]) for d in data}

        def check_output(cmd, **kwargs):
            try:
                return by_func[cmd[1]]
            except KeyError:
                raise subprocess.CalledProcessError(
                    2, cmd, output='No guesses that match criteria!')

        self.patcher = patch('subprocess.check_output', new=check_output)
        self.patcher.start()



"""Internal representation of type objects."""

from typing import NamedTuple, Optional, Sequence


class AbstractType(object):
    """Abstract base class for types."""


class ClassType(AbstractType):
    """A class type, potentially generic (int, List[str], None, ...)"""

    def __init__(self, name, args=None):
        # type: (str, Optional[Sequence[AbstractType]]) -> None
        self.name = name
        if args:
            self.args = tuple(args)
        else:
            self.args = ()

    def __repr__(self):
        # type: () -> str
        if self.name == 'Tuple' and len(self.args) == 1:
            return 'Tuple[%s, ...]' % self.args[0]
        elif self.args:
            return '%s[%s]' % (self.name, ', '.join(str(arg) for arg in self.args))
        else:
            return self.name

    def __eq__(self, other):
        # type: (object) -> bool
        return isinstance(other, ClassType) and self.name == other.name and self.args == other.args

    def __hash__(self):
        # type: () -> int
        return hash((self.name, self.args))


class AnyType(AbstractType):
    """The type Any"""

    def __repr__(self):
        # type: () -> str
        return 'Any'

    def __eq__(self, other):
        # type: (object) -> bool
        return isinstance(other, AnyType)

    def __hash__(self):
        # type: () -> int
        return hash('Any')


class NoReturnType(AbstractType):
    """The type mypy_extensions.NoReturn"""

    def __repr__(self):
        # type: () -> str
        return 'NoReturn'

    def __eq__(self, other):
        # type: (object) -> bool
        return isinstance(other, NoReturnType)

    def __hash__(self):
        # type: () -> int
        return hash('NoReturn')


class TupleType(AbstractType):
    """Fixed-length tuple Tuple[x, ..., y]"""

    def __init__(self, items):
        # type: (Sequence[AbstractType]) -> None
        self.items = tuple(items)

    def __repr__(self):
        # type: () -> str
        if not self.items:
            return 'Tuple[()]'  # Special case
        return 'Tuple[%s]' % ', '.join(str(item) for item in self.items)

    def __eq__(self, other):
        # type: (object) -> bool
        return isinstance(other, TupleType) and self.items == other.items

    def __hash__(self):
        # type: () -> int
        return hash(('tuple', self.items))


class UnionType(AbstractType):
    """Union[x, ..., y]"""

    def __init__(self, items):
        # type: (Sequence[AbstractType]) -> None
        self.items = tuple(items)

    def __repr__(self):
        # type: () -> str
        items = self.items
        if len(items) == 2:
            if is_none(items[0]):
                return 'Optional[%s]' % items[1]
            elif is_none(items[1]):
                return 'Optional[%s]' % items[0]
        return 'Union[%s]' % ', '.join(str(item) for item in items)

    def __eq__(self, other):
        # type: (object) -> bool
        return isinstance(other, UnionType) and set(self.items) == set(other.items)

    def __hash__(self):
        # type: () -> int
        return hash(('union', self.items))


# Argument kind
ARG_POS = 'ARG_POS'  # Normal
ARG_STAR = 'ARG_STAR'  # *args
ARG_STARSTAR = 'ARG_STARSTAR'  # **kwargs

# Description of an argument in a signature. The kind is one of ARG_*.
Argument = NamedTuple('Argument', [('type', AbstractType), ('kind', str)])


def is_none(t):
    # type: (AbstractType) -> bool
    return isinstance(t, ClassType) and t.name == 'None'


def is_optional(t):
    # type: (AbstractType) -> bool
    return (isinstance(t, UnionType)
            and len(t.items) == 2
            and any(item == ClassType('None') for item in t.items))



from __future__ import absolute_import, print_function

import inspect
import logging
import re
from collections import OrderedDict
from typing import Any, Iterable, List, Optional, Tuple, Type

from typeright.docs.parsers import Arg

if False:
    from typing import *

YIELDS_ERROR = "'Yields' is not allowed. Use 'Returns' with Iterator[], " \
               "or enable allow_yields"
NAMED_ITEMS_ERROR = 'Named results are not allowed. Use Tuple[] or ' \
                    'NamedTuple, or enable allow_named_results'
NAMED_RESULTS_REG = re.compile(r'[a-zA-Z_][a-zA-Z0-9_]*\s+\(([^)]+)\)')

_logger = None  # type: Optional[logging.Logger]

# mapping of arg -> type annotation
default_arg_types = {}  # type: Dict[str, str]
default_return_type = 'Any'


class FormatLoggingAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        line = self.extra['line'] + kwargs['extra']['line']
        return u'%s: line %s: %s' % (self.extra['file'], line, msg), kwargs


def get_deafult_logger():
    # type: () -> logging.Logger
    # let's us defer creation of the logger until basicConfig is called by
    # main()
    global _logger
    # FIXME: conditionally run basicConfig?
    if _logger is None:
        _logger = logging.getLogger(__name__)
    return _logger


def _cleandoc(docstring):
    # type: (str) -> str
    return inspect.cleandoc(docstring).replace('`', '')


def compile(*regexs):
    # type: (str) -> List[Pattern]
    return [re.compile(s) for s in regexs]


class DocstringFormat(object):
    name = None  # type: str
    sections = None  # type: List[Pattern]

    def __init__(self, line, filename='<string>', logger=None, options=None):
        # type: (int, Any, Optional[logging.Logger], Any) -> None
        """
        Parameters
        ----------
        line : int
            start line of the docstring
        logger : Optional[logging.Logger]
        """
        assert logger is None
        self.logger = FormatLoggingAdapter(
            logger or get_deafult_logger(),
            extra={
                'file': filename,
                'line': line,
                'column': 0
            }
        )
        self.line = line
        self.filename = filename
        self.options = options or {}

    def warning(self, message, line):
        # type: (str, int) -> None
        """
        Parameters
        ----------
        message : str
        line : int
        """
        extra = {
            'line': line,
        }
        self.logger.warning(message, extra=extra)

    def error(self, message, line):
        # type: (str, int) -> None
        """
        Parameters
        ----------
        message : str
        line : int
        """
        extra = {
            'line': line,
        }
        self.logger.error(message, extra=extra)

    @classmethod
    def matches(cls, docstring):
        # type: (str) -> bool
        """
        Return whether `docstring` is compatible with this format.

        Parameters
        ----------
        docstring : str

        Returns
        -------
        bool
        """
        return any(s.search(docstring) for s in cls.sections)

    def _cast_pararms(self, params):
        # type: (Optional[Iterable[Tuple[str, Arg]]]) -> OrderedDict[str, Arg]
        """
        Parameters
        ----------
        params : Optional[Iterable[Tuple[str, Arg]]]

        Returns
        -------
        OrderedDict[str, Arg]
        """
        if params is not None:
            return OrderedDict(params)
        else:
            return OrderedDict()

    def _cast_returns(self, returns):
        # type: (Optional[List[Arg]]) -> Optional[Arg]
        """
        Parameters
        ----------
        returns : Optional[List[Arg]]

        Returns
        -------
        Optional[Arg]
        """
        if not returns:
            return None
        elif len(returns) == 1:
            return returns[0]
        else:
            if self.options.get('allow_named_results', True):
                return Arg('Tuple[%s]' % ', '.join([x.type for x in returns
                                                    if x.type is not None]),
                           returns[0].line)
            else:
                self.warning(NAMED_ITEMS_ERROR, returns[0].line)
        return None

    def _cast_yields(self, yields):
        # type: (Optional[List[Arg]]) -> Optional[Arg]
        """
        Parameters
        ----------
        yields : Optional[List[Arg]]

        Returns
        -------
        Optional[Arg]
        """
        if not yields:
            return None

        if self.options.get('allow_yields', True):
            result = self._cast_returns(yields)
            if result:
                return Arg('Iterator[%s]' % result.type, result.line)
        else:
            self.warning(YIELDS_ERROR, yields[0].line)
        return None

    def _parse(self, docstring):
        # type: (str) -> Tuple[Optional[List[Tuple[str, Arg]]], Optional[List[Arg]], Optional[List[Arg]]]
        """
        Low-level parsing

        Parameters
        ----------
        docstring : str

        Returns
        -------
        params : Optional[List[Tuple[str, Arg]]]
            list of parsed parameters
        returns : Optional[List[Arg]]
            list of parsed return fields (list is greater than one if
            return value was documented as a tuple)
        yields : Optional[List[Arg]]
            list of parsed yield fields (list is greater than one if
            return value was documented as a tuple)
        """
        raise NotImplementedError

    def parse(self, docstring):
        # type: (str) -> Tuple[OrderedDict[str, Arg], Optional[Arg]]
        """
        Parameters
        ----------
        docstring : str

        Returns
        -------
        params : OrderedDict[str, Arg]
        result : Optional[Arg]
        """
        params, returns, yields = self._parse(_cleandoc(docstring))
        if returns and yields:
            self.warning("types found for both return and yield",
                         returns[0].line)
        if yields:
            result = self._cast_yields(yields)
        elif returns:
            result = self._cast_returns(returns)
        else:
            result = None
        return self._cast_pararms(params), result


class RestFormat(DocstringFormat):

    name = 'rest'
    sections = compile(
        r'(\n|^):param ',
        r'(\n|^):rtype:',
        r'(\n|^):Yields:'
    )

    @property
    def parser_class(self):
        from typeright.docs.parsers.rest import RestDocstring
        return RestDocstring

    def _parse(self, docstring):
        return self.parser_class(docstring, self).parse()


class NumpyFormat(DocstringFormat):
    name = 'numpy'
    sections = compile(
        r'(\n|^)Parameters\n----------\n',
        r'(\n|^)Returns\n-------\n',
        r'(\n|^)Yields\n------\n'
    )

    @property
    def parser_class(self):
        from typeright.docs.parsers.other import NumpyDocstring
        return NumpyDocstring

    def _parse(self, docstring):
        return self.parser_class(docstring).parse()


class GoogleFormat(DocstringFormat):
    name = 'google'
    sections = compile(
        r'(\n|^)Args:\n',
        r'(\n|^)Returns:\n',
        r'(\n|^)Yields:\n'
    )

    @property
    def parser_class(self):
        from typeright.docs.parsers.other import GoogleDocstring
        return GoogleDocstring

    def _parse(self, docstring):
        return self.parser_class(docstring).parse()


formats = [NumpyFormat, GoogleFormat, RestFormat]  # type: List[Type[DocstringFormat]]
format_map = {f.name: f for f in formats}  # type: Dict[str, Type[DocstringFormat]]


def guess_format(docstring):
    # type: (str) -> Optional[Type[DocstringFormat]]
    """
    Convert the passed docstring to reStructuredText format.

    Parameters
    ----------
    docstring : str
        the docstring to convert

    Returns
    -------
    Optional[Type[DocstringFormat]]
    """
    docstring = inspect.cleandoc(docstring)
    for format in formats:
        if format.matches(docstring):
            return format
    return None


def parse_docstring(docstring, line=0, filename='<string>', logger=None,
                    format_name=None, options=None):
    # type: (str, int, Any, Optional[logging.Logger], Optional[str], Any) -> Tuple[OrderedDict[str, Arg], Optional[Arg]]
    """
    Parse the passed docstring.

    The OrderedDict holding parsed parameters may be sparse.

    Parameters
    ----------
    docstring : str
    line : int
        start line of the docstring
    logger : Optional[logging.Logger]
    format_name : Optional[str]

    Returns
    -------
    params : OrderedDict[str, Arg]
    results : Optional[Arg]
    """
    if format_name is None or format_name == 'auto':
        format_cls = guess_format(docstring)
        if format_cls is None:
            format_cls = RestFormat
    else:
        format_cls = format_map[format_name]
    format = format_cls(line, filename=filename, logger=logger,
                        options=options)
    return format.parse(docstring)



from __future__ import absolute_import, print_function

import re
from lib2to3.pgen2 import token
from lib2to3.pytree import Leaf, Node

from .base import BaseFixAnnotate


class FixAnnotateAny(BaseFixAnnotate):
    """Fixer that inserts Any for all types"""

    def make_annotation(self, node, results):
        name = results['name']
        assert isinstance(name, Leaf), repr(name)
        assert name.type == token.NAME, repr(name)
        decorators = self.get_decorators(node)
        is_method = self.is_method(node)
        if name.value == '__init__' or not self.has_return_exprs(node):
            restype = 'None'
        else:
            restype = 'Any'
        args = results.get('args')
        argtypes = []
        if isinstance(args, Node):
            children = args.children
        elif isinstance(args, Leaf):
            children = [args]
        else:
            children = []
        # Interpret children according to the following grammar:
        # (('*'|'**')? NAME ['=' expr] ','?)*
        stars = inferred_type = ''
        in_default = False
        at_start = True
        for child in children:
            if isinstance(child, Leaf):
                if child.value in ('*', '**'):
                    stars += child.value
                elif child.type == token.NAME and not in_default:
                    if not is_method or not at_start or 'staticmethod' in decorators:
                        inferred_type = 'Any'
                    else:
                        # Always skip the first argument if it's named 'self'.
                        # Always skip the first argument of a class method.
                        if child.value == 'self' or 'classmethod' in decorators:
                            pass
                        else:
                            inferred_type = 'Any'
                elif child.value == '=':
                    in_default = True
                elif in_default and child.value != ',':
                    if child.type == token.NUMBER:
                        if re.match(r'\d+[lL]?$', child.value):
                            inferred_type = 'int'
                        else:
                            inferred_type = 'float'  # TODO: complex?
                    elif child.type == token.STRING:
                        if child.value.startswith(('u', 'U')):
                            inferred_type = 'unicode'
                        else:
                            inferred_type = 'str'
                    elif child.type == token.NAME and child.value in ('True', 'False'):
                        inferred_type = 'bool'
                elif child.value == ',':
                    if inferred_type:
                        argtypes.append(stars + inferred_type)
                    # Reset
                    stars = inferred_type = ''
                    in_default = False
                    at_start = False
        if inferred_type:
            argtypes.append(stars + inferred_type)
        return argtypes, restype



import unittest

from typeright.annotations.types import (AnyType, ClassType, TupleType,
                                          UnionType)


class TestTypes(unittest.TestCase):
    def test_instance_str(self):
        # type: () -> None
        assert str(ClassType('int')) == 'int'
        assert str(ClassType('List', [ClassType('int')])) == 'List[int]'
        assert str(ClassType('Dict', [ClassType('int'),
                                      ClassType('str')])) == 'Dict[int, str]'

    def test_any_type_str(self):
        # type: () -> None
        assert str(AnyType()) == 'Any'

    def test_tuple_type_str(self):
        # type: () -> None
        assert str(TupleType([ClassType('int')])) == 'Tuple[int]'
        assert str(TupleType([ClassType('int'),
                              ClassType('str')])) == 'Tuple[int, str]'
        assert str(TupleType([])) == 'Tuple[()]'

    def test_union_type_str(Self):
        # type: () -> None
        assert str(UnionType([ClassType('int'), ClassType('str')])) == 'Union[int, str]'

    def test_optional(Self):
        # type: () -> None
        assert str(UnionType([ClassType('str'), ClassType('None')])) == 'Optional[str]'
        assert str(UnionType([ClassType('None'), ClassType('str')])) == 'Optional[str]'
        assert str(UnionType([ClassType('None'), ClassType('str'),
                              ClassType('int')])) == 'Union[None, str, int]'

    def test_uniform_tuple_str(self):
        # type: () -> None
        assert str(ClassType('Tuple', [ClassType('int')])) == 'Tuple[int, ...]'



# flake8: noqa
# Our flake extension misfires on type comments in strings below.

import sys
import unittest

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from lib2to3.tests.test_fixers import FixerTestCase


class AnnotateFromSignatureTestCase(FixerTestCase):

    def setTestData(self, data):
        raise NotImplementedError

    def test_basic(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["Foo", "Bar"],
                  "return_type": "Any"},
              }])
        a = """\
            class Foo: pass
            class Bar: pass
            def nop(foo, bar):
                return 42
            """
        b = """\
            from typing import Any
            class Foo: pass
            class Bar: pass
            def nop(foo: Foo, bar: Bar) -> Any:
                return 42
            """
        self.check(a, b)

    def test_keyword_only_argument(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["Foo", "Bar"],
                  "return_type": "Any"},
              }])
        a = """\
            class Foo: pass
            class Bar: pass
            def nop(foo, *, bar):
                return 42
            """
        b = """\
            from typing import Any
            class Foo: pass
            class Bar: pass
            def nop(foo: Foo, *, bar: Bar) -> Any:
                return 42
            """
        self.check(a, b)

    def test_add_typing_import(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              # Check with and without 'typing.' prefix
              "signature": {
                  "arg_types": ["List[typing.AnyStr]", "Callable[[], int]"],
                  "return_type": "object"},
              }])
        a = """\
            def nop(foo, bar):
                return 42
            """
        b = """\
            from typing import AnyStr
            from typing import Callable
            from typing import List
            def nop(foo: List[AnyStr], bar: Callable[[], int]) -> object:
                return 42
            """
        self.check(a, b)

    def test_add_other_import(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "mod1.py",
              "line": 1,
              "signature": {
                  "arg_types": ["mod1.MyClass", "mod2.OtherClass"],
                  "return_type": "mod3.AnotherClass"},
              }])
        a = """\
            def nop(foo, bar):
                return AnotherClass()
            class MyClass: pass
            """
        b = """\
            from typing import TYPE_CHECKING
            if TYPE_CHECKING:
                from mod2 import OtherClass
                from mod3 import AnotherClass
            def nop(foo: MyClass, bar: OtherClass) -> AnotherClass:
                return AnotherClass()
            class MyClass: pass
            """
        self.check(a, b)

    def test_add_kwds(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "object"},
              }])
        a = """\
            def nop(foo, **kwds):
                return 42
            """
        b = """\
            from typing import Any
            def nop(foo: int, **kwds: Any) -> object:
                return 42
            """
        self.check(a, b)

    def test_dont_add_kwds(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int", "**AnyStr"],
                  "return_type": "object"},
              }])
        a = """\
            def nop(foo, **kwds):
                return 42
            """
        b = """\
            from typing import AnyStr
            def nop(foo: int, **kwds: AnyStr) -> object:
                return 42
            """
        self.check(a, b)

    def test_add_varargs(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "object"},
              }])
        a = """\
            def nop(foo, *args):
                return 42
            """
        b = """\
            from typing import Any
            def nop(foo: int, *args: Any) -> object:
                return 42
            """
        self.check(a, b)

    def test_dont_add_varargs(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int", "*int"],
                  "return_type": "object"},
              }])
        a = """\
            def nop(foo, *args):
                return 42
            """
        b = """\
            def nop(foo: int, *args: int) -> object:
                return 42
            """
        self.check(a, b)

    def test_return_expr_not_none(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "None"},
              }])
        a = """\
            def nop():
                return 0
            """
        b = """\
            from typing import Any
            from typing import Optional
            def nop() -> Optional[Any]:
                return 0
            """
        self.check(a, b)

    def test_return_expr_none(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "None"},
              }])
        a = """\
            def nop():
                return
            """
        b = """\
            def nop() -> None:
                return
            """
        self.check(a, b)

    def test_generator_optional(self):
        self.setTestData(
            [{"func_name": "gen",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "Optional[int]"},
              }])
        a = """\
            def gen():
                yield 42
            """
        b = """\
            from typing import Iterator
            def gen() -> Iterator[int]:
                yield 42
            """
        self.check(a, b)

    def test_generator_plain(self):
        self.setTestData(
            [{"func_name": "gen",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            def gen():
                yield 42
            """
        b = """\
            from typing import Iterator
            def gen() -> Iterator[int]:
                yield 42
            """
        self.check(a, b)

    def test_not_generator(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            def nop():
                def gen():
                    yield 42
            """
        b = """\
            def nop() -> int:
                def gen():
                    yield 42
            """
        self.check(a, b)

    def test_add_self(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            def nop(self):
                pass
            """
        b = """\
            from typing import Any
            def nop(self: Any) -> int:
                pass
            """
        self.check(a, b)

    def test_dont_add_self(self):
        self.setTestData(
            [{"func_name": "C.nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            class C:
                def nop(self):
                    pass
            """
        b = """\
            class C:
                def nop(self) -> int:
                    pass
            """
        self.check(a, b)

    def test_too_many_types(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "int"},
              }])
        a = """\
            def nop():
                pass
            """
        self.warns(a, a, "source has 0 args, annotation has 1 -- skipping", unchanged=True)

    def test_too_few_types(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": [],
                  "return_type": "int"},
              }])
        a = """\
            def nop(a):
                pass
            """
        self.warns(a, a, "source has 1 args, annotation has 0 -- skipping", unchanged=True)

    def test_classmethod(self):
        # Class method names currently are returned without class name
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "int"}
              }])
        a = """\
            class C:
                @classmethod
                def nop(cls, a):
                    return a
            """
        b = """\
            class C:
                @classmethod
                def nop(cls, a: int) -> int:
                    return a
            """
        self.check(a, b)

    def test_staticmethod(self):
        # Static method names currently are returned without class name
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "int"}
              }])
        a = """\
            class C:
                @staticmethod
                def nop(a):
                    return a
            """
        b = """\
            class C:
                @staticmethod
                def nop(a: int) -> int:
                    return a
            """
        self.check(a, b)

    def test_long_form(self):
        self.maxDiff = None
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int", "int", "int",
                                "str", "str", "str",
                                "Optional[bool]", "Union[int, str]", "*Any"],
                  "return_type": "int"},
              }])
        a = """\
            def nop(a, b, c,  # some comment
                    d, e, f,  # multi-line
                              # comment
                    g=None, h=0, *args):
                return 0
            """
        b = """\
            from typing import Any
            from typing import Optional
            from typing import Union
            def nop(a: int, b: int, c: int,  # some comment
                    d: str, e: str, f: str,  # multi-line
                              # comment
                    g: Optional[bool] = None, h: Union[int, str] = 0, *args: Any) -> int:
                return 0
            """
        self.check(a, b)

    def test_long_form_method(self):
        self.maxDiff = None
        self.setTestData(
            [{"func_name": "C.nop",
              "path": "<string>",
              "line": 2,
              "signature": {
                  "arg_types": ["int", "int", "int",
                                "str", "str", "str",
                                "Optional[bool]", "Union[int, str]", "*Any"],
                  "return_type": "int"},
              }])
        a = """\
            class C:
                def nop(self, a, b, c,  # some comment
                              d, e, f,  # multi-line
                                        # comment
                              g=None, h=0, *args):
                    return 0
            """
        b = """\
            from typing import Any
            from typing import Optional
            from typing import Union
            class C:
                def nop(self, a: int, b: int, c: int,  # some comment
                              d: str, e: str, f: str,  # multi-line
                                        # comment
                              g: Optional[bool] = None, h: Union[int, str] = 0, *args: Any) -> int:
                    return 0
            """
        self.check(a, b)

    def test_long_form_classmethod(self):
        self.maxDiff = None
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["int", "int", "int",
                                "str", "str", "str",
                                "Optional[bool]", "Union[int, str]", "*Any"],
                  "return_type": "int"},
              }])
        a = """\
            class C:
                @classmethod
                def nop(cls, a, b, c,  # some comment
                        d, e, f,
                        g=None, h=0, *args):
                    return 0
            """
        b = """\
            from typing import Any
            from typing import Optional
            from typing import Union
            class C:
                @classmethod
                def nop(cls, a: int, b: int, c: int,  # some comment
                        d: str, e: str, f: str,
                        g: Optional[bool] = None, h: Union[int, str] = 0, *args: Any) -> int:
                    return 0
            """
        self.check(a, b)
        # Do the same test for staticmethod
        a = a.replace('classmethod', 'staticmethod')
        b = b.replace('classmethod', 'staticmethod')
        self.check(a, b)

    def test_long_form_trailing_comma(self):
        self.maxDiff = None
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ["int", "int", "int",
                                "str", "str", "str",
                                "Optional[bool]", "Union[int, str]"],
                  "return_type": "int"},
              }])
        a = """\
            def nop(a, b, c,  # some comment
                    d, e, f,
                    g=None, h=0):
                return 0
            """
        b = """\
            from typing import Optional
            from typing import Union
            def nop(a: int, b: int, c: int,  # some comment
                    d: str, e: str, f: str,
                    g: Optional[bool] = None, h: Union[int, str] = 0) -> int:
                return 0
            """
        self.check(a, b)

    def test_one_liner(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["int"],
                  "return_type": "int"},
              }])
        a = """\
            def nop(a):   return a
            """
        b = """\
            def nop(a: int) -> int:   return a
            """
        self.check(a, b)

    def test_variadic(self):
        self.setTestData(
            [{"func_name": "nop",
              "path": "<string>",
              "line": 1,
              "signature": {
                  "arg_types": ["Tuple[int, ...]"],
                  "return_type": "int"},
              }])
        a = """\
            def nop(a):   return 0
            """
        b = """\
            from typing import Tuple
            def nop(a: Tuple[int, ...]) -> int:   return 0
            """
        self.check(a, b)

    @unittest.skipIf(sys.version_info < (3, 5), 'async not supported on old python')
    def test_nested_class_async_func(self):
        self.setTestData(
            [{"func_name": "A.B.foo",
              "path": "<string>",
              "line": 3,
              "signature": {
                  "arg_types": ['str'],
                  "return_type": "int"},
              }])
        a = """\
            class A:
                class B:
                    async def foo(x):
                        return 42
            """
        b = """\
            class A:
                class B:
                    async def foo(x: str) -> int:
                        return 42
            """
        self.check(a, b)

    @patch('typeright.fixes.fix_annotate_json.BaseFixAnnotateFromSignature.set_filename')
    def test_set_filename(self, mocked_set_filename):
        self.filename = "/path/to/fileA.py"
        # trigger the fixer to run, with no expected changes
        self.unchanged("")
        mocked_set_filename.assert_called_with("/path/to/fileA.py")

        self.filename = "/path/to/fileB.py"
        # trigger the fixer to run, with no expected changes
        self.unchanged("")
        mocked_set_filename.assert_called_with("/path/to/fileB.py")



"""
typeright specific fixer utility functions
"""

from lib2to3.fixer_util import (FromImport, Leaf, Newline, Node,
                                does_tree_import, find_root, is_import,
                                make_suite, syms, token)
from lib2to3.pytree import Base
from typing import (TYPE_CHECKING, Any, Callable, Dict, Generator, NamedTuple,
                    Optional, Set, Tuple, Union)

from kids.cache import cache

# ------------- caching helpers --------------- #


class HashableNode(object):
    """
    Hashable wrapper for node and leaf objects in order to cache
    """

    def __init__(self, node):
        # type: (Union[Node, Leaf]) -> None
        """
        Parameters
        -----------
        node : Union[Node, Leaf]
        """
        self.node = node

    def __hash__(self):
        return hash((self.node.get_lineno(),
                     self.node.depth(),
                     self.node.__str__()))


def wrap_cachable(func):
    """
    Converts the CST node objects of a function into HashableNodes to allow the decorated
    function to be cached (provided that all other arguments are hashable).
    """
    def inner(*args, **kwargs):
        args, kwargs = _transform_by_type(HashableNode, Base,
                                          *args, **kwargs)

        return func(*args, **kwargs)
    return inner


def unwrap_cachable(func):
    """
    Converts any HashableNodes in the argument list of a function into their standard node
    counterparts.
    """
    def inner(*args, **kwargs):
        args, kwargs = _transform_by_type(lambda hashable: hashable.node, HashableNode,
                                          *args, **kwargs)

        return func(*args, **kwargs)
    return inner


def node_cache(func):
    """
    Catch-all decorator to handle caching nodes of the CST
    """
    return wrap_cachable(cache(unwrap_cachable(func)))


# ------------------ other decorators ------------------ #


def force_root_args(func):
    """
    Given any function with a node argument, this decorator converts the node into its root parent.
    If a function has this decorator, all nodes with the same root will be treated the same.

    In our codebase, we often have a singular node available, but want to access the root of that
    node in some function. If we want to cache such a function, we must convert the node into root
    before we cache rather than within the function. This is the motivation behind this decorator.
    """
    def inner(*args, **kwargs):
        args, kwargs = _transform_by_type(find_root, Base,
                                          *args, **kwargs)

        return func(*args, **kwargs)
    return inner


def _transform_by_type(transformer, type, *args, **kwargs):
    """
    Convenience function for our decorators
    """
    new_args = (transformer(arg) if isinstance(arg, type) else arg for arg in args)
    new_kwargs = {k: transformer(v) if isinstance(v, type) else v for k, v in kwargs}

    return new_args, new_kwargs


# ------------------- utils ---------------------- #


def type_by_import_stmt(package, name, node):
    # type: (str, str, Node) -> Optional[str]
    """
    Return name that needs to be annotated based on available imports

    Parameters
    -------------
    package : str
    name : str
        Name of type being imported
    node : Node

    Returns
    -------------
    Optional[str]
        Returns valid name for type if current import statement exists, or None is there is no
        import statement
    """
    root = find_root(node)

    import_info = find_import_info(package, name, root)

    if import_info:
        # name is being imported directly in the form of
        # from <package> import <name>
        return import_info.binding

    split_path = package.rsplit('.', 1)
    if len(split_path) > 1:
        # We have a package of the form 'pkg.mod'
        pkg = split_path[0]
        mod = split_path[1]
    else:
        # We have a package of the form 'mod'
        pkg = ''
        mod = package
    module_import_info = find_import_info(pkg, mod, root)

    if module_import_info:
        return '.'.join((module_import_info.binding, name))

    return None


def is_import_stmt(node):
    return (node.type == syms.simple_stmt and node.children and
            is_import(node.children[0]))


def _generate_import_node(package, name, prefix=""):

    def DottedName(name, prefix=""):
        split = name.rsplit('.')
        if len(split) > 1:
            # Reconstruct the dotted name as a list of leaves
            leftmost_name = Leaf(token.NAME, split[0])
            children = [leftmost_name]
            for entry in split[1:]:
                next_name = [Leaf(token.DOT, '.'), Leaf(token.NAME, entry)]
                children.extend(next_name)
            return Node(syms.dotted_name, children, prefix=prefix)
        return Leaf(token.NAME, name, prefix=prefix)

    if not package:
        import_ = Node(syms.import_name, [
            Leaf(token.NAME, "import", prefix=prefix),
            DottedName(name, prefix=" ")
        ])
    else:
        import_ = Node(syms.import_from, [
            Leaf(token.NAME, "from", prefix=prefix),
            DottedName(package, prefix=" "),
            Leaf(token.NAME, "import", prefix=" "),
            Leaf(token.NAME, name, prefix=" "),
        ])

    return import_


def _get_bottom_of_imports(root):
    # type: (Node) -> int
    """
    Returns the bottom of global import block if there is one. If there is none it returns the
    bottom of the file docstring. If there is no docstring it returns the top of the file.

    Parameters
    -----------
    root : Node

    Return
    -----------
    int
    """
    # figure out where to insert the new import.  First try to find
    # the first import and then skip to the last one.
    insert_pos = offset = 0
    for idx, node in enumerate(root.children):
        if not is_import_stmt(node):
            continue
        for offset, node2 in enumerate(root.children[idx:]):
            if not is_import_stmt(node2):
                break
        insert_pos = idx + offset
        break

    # if there are no imports where we can insert, find the docstring.
    # if that also fails, we stick to the beginning of the file
    if insert_pos == 0:
        for idx, node in enumerate(root.children):
            if (node.type == syms.simple_stmt and node.children and
                    node.children[0].type == token.STRING):
                insert_pos = idx + 1
                break

    return insert_pos


def create_import(package, name, node):
    # type: (str, str, Node) -> None
    """
    Create import statement of the form `from <package> import <name>`

    Parameters
    -------------
    package : str
    name : str
        Name of type being imported
    node : Node
    """

    root = find_root(node)

    insert_pos = _get_bottom_of_imports(root)

    import_ = _generate_import_node(package, name)

    children = [import_, Newline()]
    root.insert_child(insert_pos, Node(syms.simple_stmt, children))


# -------------- TYPE_CHECKING LOGIC ------------------ #


def _new_type_check_with_import(package, name, root, insert_pos):
    # type: (Optional[str], str, Node, int) -> None
    """
    Inserts a new TYPE_CHECKING block containing a new import statement for package and name

    Parameters
    -----------
    package : Optional[str]
    name : str
    root : Node
    insert_pos : int
    """
    # [Grammar]
    # if_stmt: 'if' namedexpr_test ':' suite ('elif' namedexpr_test ':' suite)* ['else' ':' suite]
    type_check_node = Node(syms.if_stmt,
                           [Leaf(token.NAME, 'if'),
                            Leaf(token.NAME, 'TYPE_CHECKING', prefix=" "),
                            Leaf(token.COLON, ':'),
                            # [Grammar]
                            # suite: simple_stmt | NEWLINE INDENT stmt+ DEDENT
                            Node(syms.suite, [Leaf(token.NEWLINE, '\n'),
                                              Leaf(token.INDENT, '    '),
                                              Node(syms.simple_stmt,
                                                   [_generate_import_node(package, name), Newline()]),
                                              Leaf(token.DEDENT, '')])])

    # We can just hardcode the correct insert position since we just created the typing block
    root.insert_child(insert_pos, type_check_node)
    # Make sure to import TYPE_CHECKING just before using
    import_type_checking = [_generate_import_node('typing', 'TYPE_CHECKING'), Newline()]
    root.insert_child(insert_pos, Node(syms.simple_stmt, import_type_checking))


def create_type_checking_import(package, name, node):
    # type: (str, str, Node) -> None
    """
    Create import statement of the form `from <package> import <name>` within a TYPING_CHECK
    block

    Parameters
    -------------
    package : str
    name : str
        Name of type being imported
    node : Node
    """

    def is_type_checking_decl(node):
        # [Grammar]
        # if_stmt: 'if' namedexpr_test ':' suite ('elif' namedexpr_test ':' suite)* ['else' ':' suite]
        if not node.type == syms.if_stmt:
            return False
        stmt = str(node.children[1]).strip()
        if stmt in ('typing.TYPE_CHECKING', 'TYPE_CHECKING'):
            return True

        return False

    root = find_root(node)

    # figure out where to insert the new import.  First try to find
    # the first import and then skip to the last one.
    type_checking_suite = None
    parent = root
    for idx, node in enumerate(root.children):
        if not is_type_checking_decl(node):
            continue

        type_checking_suite = node.children[3]
        parent = type_checking_suite
        insert_pos = len(type_checking_suite.children) - 1
        break

    if type_checking_suite is None:
        # Generate a new TYPE_CHECKING block at the bottom of the current import block and return
        insert_pos = _get_bottom_of_imports(root)
        _new_type_check_with_import(package, name, root, insert_pos)
        return

    import_ = _generate_import_node(package, name, prefix="    ")
    children = [import_, Newline()]

    parent.insert_child(insert_pos, Node(syms.simple_stmt, children))


@force_root_args
@node_cache
def get_unprotected_imports(root):
    """
    Returns all imports of a syntax tree at the global level. So any imports inside a TYPE_CHECKING
    block or defined inside of a method or class are not included. Intended to be used once, before
    any editing of the tree, so we can cache for efficiency.
    """
    imports = set([])  # type: Set[str]
    for child in root.children:
        if is_import_stmt(child):
            ret = get_import_info(child.children[0])
            imports = set(ret.imports.keys())

    return imports


def touch_import(package, name, node):
    # type: (str, str, Node) -> str
    """
    Almost the same as lib2to3/fixer_util's touch_import, but broken into helper functions.
    The difference between the lib2to3's function and ours is that this function assumes a
    statement of the form:
        `import mod`
    is a sufficient import for type `mod.Name`.

    To not obfuscate which type of import statement currently exists (and by proxy which form the
    typename needs to be written in to be valid), we return the valid form of the typename.

    So for the example above, touch_import would return
        `mod.Name`
    but if instead there was an import statement of the form
        `from mod import Name`
    then this function would simply return
        `Name`

    Parameters
    -------------
    package : str
    name : str
        Name of the type being imported
    node : Node

    Return
    -------------
    str
        The typename that will be defined by either the already existing or created import
        statement
    """
    result = type_by_import_stmt(package, name, node)

    if result is not None:
        # If there is already an import statement
        return result

    create_import(package, name, node)
    return name


_block_syms = {syms.funcdef, syms.classdef, syms.trailer}


def _find(name, node):
    nodes = [node]
    while nodes:
        node = nodes.pop()
        if node.type > 256 and node.type not in _block_syms:
            nodes.extend(node.children)
        elif node.type == token.NAME and node.value.strip() == name:
            return node
    return None


ImportPairing = NamedTuple("ImportPairing", [
    ("entry", str),
    ("binding_name", str)
])

ImportNodeInfo = NamedTuple("ImportNodeInfo", [
    ("imports", Dict[str, Set[ImportPairing]]),
    ("node", Node)
])

ImportInfo = NamedTuple("ImportInfo", [
    ("entry", str),
    ("package", str),
    ("binding", str)
])


def find_import_info(package, name, node):
    # type: (str, str, Node) -> Optional[ImportInfo]
    """
    Finds the import statement for <name> regardless of whether <name> is the binding name
        ex where name='a':
        import a => match AND import a as b => match BUT import b as a => None

    Parameters
    -----------
    package : str
    name : str
    node : Node

    Return
    -----------
    Optional[ImportInfo]
    """
    for child in node.children:
        if is_import(child):
            ret = get_import_info(child)
            imports = ret.imports.get(package)
            if imports is None:
                return None
            import_ = next((x for x in imports if x.entry == name), None)
            if import_:
                return ImportInfo(import_.entry, package, import_.binding_name)
        elif child.type == syms.simple_stmt:
            res = find_import_info(package, name, child)
            if res is None:
                continue
            else:
                return res
    return None


def decompose_name(node):
    """
    NOTE: Per the lib2to3 grammar:
            dotted_name: NAME ('.' NAME)*
          This means that dotted_name can be either dotted or not dotted, i.e. it's a generalized
          form of NAME. So this function will cover both cases.

    Given a dotted_name node this will return a tuple of the form (pkg, name, full_string) where
    all are str
        ex: a.b.c => (a.b, c, a.b.c)
            b.c => (b, c, b.c)
            c => (None, c, c)
    otherwise it will return None for each field
    """
    if node.type == token.NAME:
        # node is just a name, no dots
        return '', node.value, node.value

    if node.children:
        # Right most node will be the name, i.e. a.b.c = ['a','.','b','.','c']
        name_node = node.children[-1]
        package_nodes = node.children[:-2]
        name = str(name_node).strip()
        package = ''.join(str(n).strip() for n in package_nodes)
        full = ''.join(str(n).strip() for n in node.children)
        return package, name, full

    return None, None, None


@node_cache
def get_import_info(node):
    # type: (Node) -> ImportNodeInfo
    """
    If node is a valid import_stmt, this will return a named tuple for each component of the
    statement
        ex: from a.b import c as d, e  => (imports={'a.b': [(import_name='c',
                                                             binding_name='d'),
                                                            (import_name='e',
                                                             binding_name='e')]},
                                           node=node)

    Since we will regularly iterate over the list import node for their information when
    searching for binding or import matches, we cache the results of this function for
    quick and non-redundant lookups

    Parameters
    -----------
    node : Node

    Return
    -----------
    ImportNodeInfo
    """
    def handle_name(node):
        if node.type in (syms.dotted_as_name, syms.import_as_name):
            # [Grammar] dotted_as_name: dotted_name ['as' NAME]
            package = import_name = binding_name = None
            name = node.children[0]
            binding_name = str(node.children[2]).strip()
            if name.type in (syms.dotted_name, token.NAME):
                # [Grammar] dotted_name: NAME ('.' NAME)*
                # We don't need the third field since we know the alias will be the binding_name
                package, import_name, _ = decompose_name(name)
            return package, import_name, binding_name
        # If there was no "as", we know this is dotted_name
        # [Grammar] dotted_name: NAME ('.' NAME)*
        return decompose_name(node)

    package = None
    imports = dict()  # type: Dict[str, Set[ImportPairing]]
    if node.type == syms.import_name:
        # [Grammar]: import_name: 'import' dotted_as_names
        as_name = node.children[1]
        if as_name.type in (syms.dotted_as_names, syms.import_as_names):
            # [Grammar]: dotted_as_names: dotted_as_name (',' dotted_as_name)*
            as_names = [child for child in as_name.children if str(child).strip() != ',']
        else:
            as_names = [as_name]
        for child in as_names:
            package, import_name, binding_name = handle_name(child)
            imports.setdefault(package, set()).add(ImportPairing(import_name, binding_name))
    elif node.type == syms.import_from:
        # [Grammar]:
        # import_from: ('from' ('.'* dotted_name | '.'+)
        #   'import' ('*' | '(' import_as_names ')' | import_as_names))
        package = str(node.children[1]).strip()
        if str(node.children[3]).strip() == '(':
            # This means we are dealing with an import statement containing parentheses
            # ex: from a import (b, c, d)
            import_as_name = node.children[4]
        else:
            import_as_name = node.children[3]
        if import_as_name.type == syms.import_as_names:
            as_names = [child for child in import_as_name.children if str(child).strip() != ',']
        else:
            as_names = [import_as_name]
        for child in as_names:
            _, import_name, binding_name = handle_name(child)
            imports.setdefault(package, set()).add(ImportPairing(import_name, binding_name))

    return ImportNodeInfo(imports, node)



from __future__ import print_function

import argparse
import json
import logging
import os
import re
import sys
from lib2to3.main import StdoutRefactoringTool
from os.path import expanduser, expandvars
from typing import Any, Dict, List, Optional

import six.moves.configparser as configparser

from typeright.annotations.main import generate_annotations_json_string
from typeright.docs import formats
from typeright.fixes.base import BaseFixAnnotateFromSignature, crawl_up
from typeright.fixes.fix_annotate_any import FixAnnotateAny
from typeright.fixes.fix_annotate_command import FixAnnotateCommand
from typeright.fixes.fix_annotate_docs import FixAnnotateDocs
from typeright.fixes.fix_annotate_json import FixAnnotateJson

NAME = 'typeright'
CONFIG_FILES = [NAME + '.ini', 'setup.cfg']


def get_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument('files', nargs='*', metavar="FILE",
                        help="Files and directories to update with annotations")

    # json --
    json_group = parser.add_argument_group('json file options',
                                           "Read type info from a json file")
    json_group.add_argument('--type-info', metavar="FILE",
                            help="JSON input file")
    json_group.add_argument('--max-line-drift', type=int, default=5, metavar="N",
                            help="Maximum allowed line drift when inserting annotation"
                                 " (can be useful for custom codecs)")
    json_group.add_argument('--uses-signature', action='store_true',
                            help="JSON input uses a signature format")
    json_group.add_argument('-s', '--only-simple', action='store_true',
                            help="Only annotate functions with trivial types")

    # command --
    cmd_group = parser.add_argument_group('command options',
                                          "Generate type info by calling an "
                                          "external program")
    cmd_group.add_argument('--command', '-c', metavar="COMMAND",
                           help="Command to generate JSON info for a call site")

    # doc --
    doc_group = parser.add_argument_group('docstring options',
                                          "Generate type info by parsing docstrings")
    doc_group.add_argument("--doc-format",
                           choices=sorted(list(formats.format_map.keys()) + ['auto', 'off']),
                           help="Specify the docstring convention used within "
                                "files to be converted ('auto' automatically "
                                "determines the format by inspecting each docstring "
                                "but it is faster and more reliable to specify this "
                                "explicitly)")
    doc_group.add_argument("--doc-default-return-type",
                           default='Any', metavar="TYPE",
                           help="Default type to use for undocumented return "
                                "values (defaults to 'Any'")

    # any --
    any_group = parser.add_argument_group('any options')
    any_group.add_argument('-a', '--auto-any', action='store_true',
                           help="Annotate everything with 'Any'")

    # format --
    format_group = parser.add_argument_group('output format options')
    format_group.add_argument('--annotation-style', default='auto',
                              choices=['auto', 'py2', 'py3'],
                              help="Choose annotation style, py2 for Python 2 with "
                                   "comments, py3 for Python 3 with annotation "
                                   "syntax. The default will be determined by the "
                                   "version of the current python interpreter")
    format_group.add_argument('--py2-comment-style', default='auto',
                              choices=['auto', 'multi', 'single'],
                              help="Choose comment style, multi adds a comment per "
                                   "argument, single produces one type comment for "
                                   "all arguments, and auto chooses between the two "
                                   "styles based on the number of arguments and "
                                   "length of comments")

    # other --
    other_group = parser.add_argument_group('other options')
    other_group.add_argument('-p', '--print-function', action='store_true',
                             help="Assume print is a function")
    other_group.add_argument('-w', '--write', action='store_true',
                             help="Write output files")
    other_group.add_argument('-j', '--processes', type=int, default=1, metavar="N",
                             help="Use N parallel processes (default no parallelism)")
    other_group.add_argument('-v', '--verbose', action='store_true',
                             help="More verbose output")
    other_group.add_argument('-q', '--quiet', action='store_true',
                             help="Don't show diffs")
    other_group.add_argument("-o", "--output-dir", type=str, default="",
                             metavar="FOLDER",
                             help="Put output files in this directory "
                                  "instead of overwriting the input files.")
    other_group.add_argument("-W", "--write-unchanged-files", action='store_true',
                             help="Also write files even if no changes were required"
                                  " (useful with --output-dir); implies -w.")
    return parser


def load_config(arg_parser):
    # type: (argparse.ArgumentParser) -> Dict
    """Use the ArgumentParser to extract values set in a config file.

    These are used to set the defaults on the parser before calling
    parse_args(), which ensures that arguments specified on the command line
    have a higher priority than those set in the config file.
    """
    SPLIT = re.compile('[,\n]+')
    paths = [expanduser(expandvars(path)) for path in CONFIG_FILES]
    config_parser = configparser.RawConfigParser()
    config_parser.read(paths)
    defaults = {}
    if config_parser.has_section(NAME):
        for action in arg_parser._actions:
            if action.dest in defaults:
                # some arguments may refer to the same destination
                continue
            try:
                if isinstance(action, argparse._HelpAction):
                    continue
                elif isinstance(action, (argparse._StoreFalseAction,
                                         argparse._StoreTrueAction,
                                         argparse._StoreAction,
                                         argparse._StoreConstAction)):
                    if action.type is int:
                        val = config_parser.getint(NAME, action.dest)  # type: Any
                    elif action.type is bool:
                        val = config_parser.getboolean(NAME, action.dest)
                    elif action.type is float:
                        val = config_parser.getfloat(NAME, action.dest)
                    else:
                        val = config_parser.get(NAME, action.dest)
                        if action.nargs in {'*', '+'}:
                            val = [x for x in SPLIT.split(val) if x]
                else:
                    raise TypeError(action)
            except configparser.NoOptionError:
                continue
            defaults[action.dest] = val

    return defaults


def _main(args_override=None):
    # type: (Optional[List[str]]) -> List[str]

    parser = get_parser()
    defaults = load_config(parser)
    parser.set_defaults(**defaults)
    # Parse command line.
    args = parser.parse_args(args_override)

    annotation_style = args.annotation_style
    if annotation_style == 'auto':
        annotation_style = 'py%d' % sys.version_info[0]

    # Set up logging handler.
    level = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(format='%(message)s', level=level)

    input_base_dir = crawl_up(os.path.abspath(args.files[0]))[0]
    if args.output_dir:
        input_base_dir = input_base_dir.rstrip(os.sep)
        logging.info('Output in %r will mirror the input directory %r layout.',
                     args.output_dir, input_base_dir)

    if args.write_unchanged_files:
        if not args.write:
            logging.warn("--write-unchanged-files/-W implies -w.")
        args.write = True

    fixers = []  # type: List[str]

    def add_fixer(fixer_cls):
        fixer_cls.run_order = BaseFixAnnotateFromSignature.run_order + len(fixers)
        fixers.append(fixer_cls.__module__)

    options = {
        'annotation_style': annotation_style,
        'comment_style': args.py2_comment_style,
    }

    if args.type_info:
        # Produce nice error message if type_info.json not found.
        try:
            with open(args.type_info) as f:
                contents = f.read()
        except IOError as err:
            sys.exit("Can't open type info file: %s" % err)

        # Run pass 2 with output into a variable.
        if args.uses_signature:
            data = json.loads(contents)  # type: List[Any]
        else:
            data = generate_annotations_json_string(
                args.type_info,
                only_simple=args.only_simple)

        # FIXME: in multi-process mode it may be faster to pass the json
        #  file name and re-read it, rather than copy the data into each process
        options['type_info'] = data
        options['top_dir'] = input_base_dir
        add_fixer(FixAnnotateJson)

    if args.command:
        options['command'] = args.command
        add_fixer(FixAnnotateCommand)

    if args.doc_format not in {None, 'off'}:
        options['doc_format'] = args.doc_format
        options['doc_default_return_type'] = args.doc_default_return_type
        add_fixer(FixAnnotateDocs)

    if args.auto_any:
        add_fixer(FixAnnotateAny)

    flags = {
        'write_unchanged_files': args.write_unchanged_files,
        'print_function': args.print_function,
        'typeright': options,
    }
    rt = StdoutRefactoringTool(
        fixers=fixers,
        options=flags,
        explicit=fixers,
        nobackups=True,
        show_diffs=not args.quiet,
        input_base_dir=input_base_dir,
        output_dir=args.output_dir,
    )
    if not rt.errors:
        with BaseFixAnnotateFromSignature.max_line_drift_set(args.max_line_drift):
            rt.refactor(args.files, write=args.write, num_processes=args.processes)
        if args.processes == 1:
            rt.summarize()
        else:
            logging.info("(In multi-process per-file warnings are lost)")
    if not args.write:
        logging.info("NOTE: this was a dry run; use -w to write files")
    return rt.errors


def main(args=None):
    errors = _main(args)
    # Return error status (0 if rt.errors is zero)
    return int(bool(errors))


if __name__ == '__main__':
    sys.exit(main())



import os
import tempfile
import unittest
from typing import List, Optional, Tuple

from typeright.annotations.parse import (ParseError, parse_json,
                                          parse_type_comment, tokenize)
from typeright.annotations.types import (ARG_POS, ARG_STAR, ARG_STARSTAR,
                                          AbstractType, AnyType, Argument,
                                          ClassType, NoReturnType, TupleType,
                                          UnionType)


class TestParseError(unittest.TestCase):
    def test_str_conversion(self):
        # type: () -> None
        assert str(ParseError('(int -> str')) == 'Invalid type comment: (int -> str'


class TestParseJson(unittest.TestCase):
    def test_parse_json(self):
        # type: () -> None
        data = """
        [
            {
                "path": "pkg/thing.py",
                "line": 422,
                "func_name": "my_function",
                "type_comments": [
                    "(int) -> None",
                    "(str) -> None"
                ],
                "samples": 3
            }
        ]
        """
        f = None
        try:
            with tempfile.NamedTemporaryFile(mode='w', delete=False) as f:
                f.write(data)
            result = parse_json(f.name)
        finally:
            if f is not None:
                os.remove(f.name)
        assert len(result) == 1
        item = result[0]
        assert item.path == 'pkg/thing.py'
        assert item.line == 422
        assert item.func_name == 'my_function'
        assert item.type_comments == ['(int) -> None',
                                      '(str) -> None']
        assert item.samples == 3


class TestTokenize(unittest.TestCase):
    def test_tokenize(self):
        # type: () -> None
        self.assert_tokenize(
            ' List[int, str] (  )-> *',
            'DottedName(List) [ DottedName(int) , DottedName(str) ] ( ) -> * End()')

    def test_special_cases(self):
        # type: () -> None
        self.assert_tokenize('dictionary-itemiterator',
                             'DottedName(Iterator) End()')
        self.assert_tokenize('dictionary-keyiterator',
                             'DottedName(Iterator) End()')
        self.assert_tokenize('dictionary-valueiterator',
                             'DottedName(Iterator) End()')
        self.assert_tokenize('foo-bar', 'DottedName(Any) End()')
        self.assert_tokenize('pytz.tzfile.Europe/Amsterdam',
                             'DottedName(datetime.tzinfo) End()')

    def assert_tokenize(self, s, expected):
        # type: (str, str) -> None
        tokens = tokenize(s)
        actual = ' '.join(str(t) for t in tokens)
        assert actual == expected


def class_arg(name, args=None):
    # type: (str, Optional[List[AbstractType]]) -> Argument
    return Argument(ClassType(name, args), ARG_POS)


def any_arg():
    # type: () -> Argument
    return Argument(AnyType(), ARG_POS)


def tuple_arg(items):
    # type: (List[AbstractType]) -> Argument
    return Argument(TupleType(items), ARG_POS)


class TestParseTypeComment(unittest.TestCase):
    def test_empty(self):
        # type: () -> None
        self.assert_type_comment('() -> None', ([], ClassType('None')))

    def test_simple_args(self):
        # type: () -> None
        self.assert_type_comment('(int) -> None', ([class_arg('int')], ClassType('None')))
        self.assert_type_comment('(int, str) -> bool', ([class_arg('int'),
                                                         class_arg('str')], ClassType('bool')))

    def test_generic(self):
        # type: () -> None
        self.assert_type_comment('(List[int]) -> Dict[str, bool]',
                                 ([class_arg('List', [ClassType('int')])],
                                  ClassType('Dict', [ClassType('str'), ClassType('bool')])))

    def test_any_and_unknown(self):
        # type: () -> None
        self.assert_type_comment('(Any) -> pyannotate_runtime.collect_types.UnknownType',
                                 ([any_arg()], AnyType()))

    def test_no_return(self):
        # type: () -> None
        self.assert_type_comment('() -> pyannotate_runtime.collect_types.NoReturnType',
                                 ([], NoReturnType()))

    def test_tuple(self):
        # type: () -> None
        self.assert_type_comment('(Tuple[]) -> Any', ([tuple_arg([])], AnyType()))
        self.assert_type_comment('(Tuple[int]) -> Any',
                                 ([tuple_arg([ClassType('int')])], AnyType()))
        self.assert_type_comment('(Tuple[int, str]) -> Any',
                                 ([tuple_arg([ClassType('int'),
                                              ClassType('str')])], AnyType()))

    def test_union(self):
        # type: () -> None
        self.assert_type_comment('(Union[int, str]) -> Any',
                                 ([Argument(UnionType([ClassType('int'),
                                                       ClassType('str')]), ARG_POS)], AnyType()))
        self.assert_type_comment('(Union[int]) -> Any',
                                 ([class_arg('int')], AnyType()))

    def test_optional(self):
        # type: () -> None
        self.assert_type_comment('(Optional[int]) -> Any',
                                 ([Argument(UnionType([ClassType('int'),
                                                       ClassType('None')]), ARG_POS)], AnyType()))

    def test_star_args(self):
        # type: () -> None
        self.assert_type_comment('(*str) -> Any',
                                 ([Argument(ClassType('str'), ARG_STAR)], AnyType()))
        self.assert_type_comment('(int, *str) -> Any',
                                 ([class_arg('int'), Argument(ClassType('str'), ARG_STAR)],
                                  AnyType()))

    def test_star_star_args(self):
        # type: () -> None
        self.assert_type_comment('(**str) -> Any',
                                 ([Argument(ClassType('str'), ARG_STARSTAR)], AnyType()))
        self.assert_type_comment('(int, *str, **bool) -> Any',
                                 ([class_arg('int'),
                                   Argument(ClassType('str'), ARG_STAR),
                                   Argument(ClassType('bool'), ARG_STARSTAR)], AnyType()))

    def test_function(self):
        # type: () -> None
        self.assert_type_comment('(function) -> Any',
                                 ([class_arg('Callable')], AnyType()))

    def test_unicode(self):
        # type: () -> None
        self.assert_type_comment('(unicode) -> Any',
                                 ([class_arg('Text')], AnyType()))

    def test_bad_annotation(self):
        # type: () -> None
        for bad in ['( -> None',
                    '()',
                    ')) -> None',
                    '() -> ',
                    '()->',
                    '() -> None x',
                    'int',
                    'int -> None',
                    '(Union[]) -> None',
                    '(List[int) -> None',
                    '(*int, *str) -> None',
                    '(*int, int) -> None',
                    '(**int, *str) -> None',
                    '(**int, str) -> None',
                    '(**int, **str) -> None']:
            with self.assertRaises(ParseError):
                parse_type_comment(bad)

    def assert_type_comment(self, comment, expected):
        # type: (str, Tuple[List[Argument], AbstractType]) -> None
        actual = parse_type_comment(comment)
        assert actual == expected



import unittest
from typing import List, Tuple

from typeright.annotations.infer import (flatten_types, infer_annotation,
                                          merge_items, remove_redundant_items)
from typeright.annotations.types import (ARG_POS, ARG_STAR, AbstractType,
                                          AnyType, ClassType, NoReturnType,
                                          TupleType, UnionType)


class TestInfer(unittest.TestCase):
    def test_simple(self):
        # type: () -> None
        self.assert_infer(['(int) -> str'], ([(ClassType('int'), ARG_POS)],
                                             ClassType('str')))

    def test_infer_union_arg(self):
        # type: () -> None
        self.assert_infer(['(int) -> None',
                           '(str) -> None'],
                          ([(UnionType([ClassType('int'),
                                        ClassType('str')]), ARG_POS)],
                           ClassType('None')))

    def test_infer_union_return(self):
        # type: () -> None
        self.assert_infer(['() -> int',
                           '() -> str'],
                          ([],
                           UnionType([ClassType('int'), ClassType('str')])))

    def test_star_arg(self):
        # type: () -> None
        self.assert_infer(['(int) -> None',
                           '(int, *bool) -> None'],
                          ([(ClassType('int'), ARG_POS),
                            (ClassType('bool'), ARG_STAR)],
                           ClassType('None')))

    def test_merge_unions(self):
        # type: () -> None
        self.assert_infer(['(Union[int, str]) -> None',
                           '(Union[str, None]) -> None'],
                          ([(UnionType([ClassType('int'),
                                        ClassType('str'),
                                        ClassType('None')]), ARG_POS)],
                           ClassType('None')))

    def test_remove_redundant_union_item(self):
        # type: () -> None
        self.assert_infer(['(str) -> None',
                           '(unicode) -> None'],
                          ([(ClassType('Text'), ARG_POS)],
                           ClassType('None')))

    def test_remove_redundant_dict_item(self):
        # type: () -> None
        self.assert_infer(['(Dict[str, Any]) -> None',
                           '(Dict[str, str]) -> None'],
                          ([(ClassType('Dict', [ClassType('str'), AnyType()]), ARG_POS)],
                           ClassType('None')))

    def test_remove_redundant_dict_item_when_simplified(self):
        # type: () -> None
        self.assert_infer(['(Dict[str, Any]) -> None',
                           '(Dict[str, Union[str, List, Dict, int]]) -> None'],
                          ([(ClassType('Dict', [ClassType('str'), AnyType()]), ARG_POS)],
                           ClassType('None')))

    def test_simplify_list_item_types(self):
        # type: () -> None
        self.assert_infer(['(List[Union[bool, int]]) -> None'],
                          ([(ClassType('List', [ClassType('int')]), ARG_POS)],
                           ClassType('None')))

    def test_simplify_potential_typed_dict(self):
        # type: () -> None
        # Fall back to Dict[x, Any] in case of a complex Dict type.
        self.assert_infer(['(Dict[str, Union[int, str]]) -> Any'],
                          ([(ClassType('Dict', [ClassType('str'), AnyType()]), ARG_POS)],
                           AnyType()))
        self.assert_infer(['(Dict[Text, Union[int, str]]) -> Any'],
                          ([(ClassType('Dict', [ClassType('Text'), AnyType()]), ARG_POS)],
                           AnyType()))
        # Not a potential TypedDict so ordinary simplification applies.
        self.assert_infer(['(Dict[str, Union[str, Text]]) -> Any'],
                          ([(ClassType('Dict', [ClassType('str'), ClassType('Text')]), ARG_POS)],
                           AnyType()))
        self.assert_infer(['(Dict[str, Union[int, None]]) -> Any'],
                          ([(ClassType('Dict', [ClassType('str'),
                                                UnionType([ClassType('int'),
                                                           ClassType('None')])]), ARG_POS)],
                           AnyType()))

    def test_simplify_multiple_empty_collections(self):
        # type: () -> None
        self.assert_infer(['() -> Tuple[List, List[x]]',
                           '() -> Tuple[List, List]'],
                          ([],
                           TupleType([ClassType('List'), ClassType('List', [ClassType('x')])])))

    def assert_infer(self, comments, expected):
        # type: (List[str], Tuple[List[Tuple[AbstractType, str]], AbstractType]) -> None
        actual = infer_annotation(comments)
        assert actual == expected

    def test_infer_ignore_mock(self):
        # type: () -> None
        self.assert_infer(['(mock.mock.Mock) -> None',
                           '(str) -> None'],
                          ([(ClassType('str'), ARG_POS)],
                           ClassType('None')))

    def test_infer_ignore_mock_fallback_to_any(self):
        # type: () -> None
        self.assert_infer(['(mock.mock.Mock) -> str',
                           '(mock.mock.Mock) -> int'],
                          ([(AnyType(), ARG_POS)],
                           UnionType([ClassType('str'), ClassType('int')])))

    def test_infer_none_argument(self):
        # type: () -> None
        self.assert_infer(['(None) -> None'],
                          ([(UnionType([ClassType('None'), AnyType()]), ARG_POS)],
                           ClassType('None')))


CT = ClassType


class TestRedundantItems(unittest.TestCase):
    def test_cannot_simplify(self):
        # type: () -> None
        for first, second in ((CT('str'), CT('int')),
                              (CT('List', [CT('int')]),
                               CT('List', [CT('str')])),
                              (CT('List'),
                               CT('Set', [CT('int')]))):
            assert remove_redundant_items([first, second]) == [first, second]
            assert remove_redundant_items([second, first]) == [second, first]

    def test_simplify_simple(self):
        # type: () -> None
        for first, second in ((CT('str'), CT('Text')),
                              (CT('bool'), CT('int')),
                              (CT('int'), CT('float'))):
            assert remove_redundant_items([first, second]) == [second]
            assert remove_redundant_items([second, first]) == [second]

    def test_simplify_multiple(self):
        # type: () -> None
        assert remove_redundant_items([CT('Text'), CT('str'), CT('bool'), CT('int'),
                                       CT('X')]) == [CT('Text'), CT('int'), CT('X')]

    def test_simplify_generics(self):
        # type: () -> None
        for first, second in ((CT('List'), CT('List', [CT('Text')])),
                              (CT('Set'), CT('Set', [CT('Text')])),
                              (CT('Dict'), CT('Dict', [CT('str'), CT('int')]))):
            assert remove_redundant_items([first, second]) == [second]


class TestMergeUnionItems(unittest.TestCase):
    def test_cannot_merge(self):
        # type: () -> None
        for first, second in ((CT('str'), CT('Text')),
                              (CT('List', [CT('int')]), CT('List', [CT('str')]))):
            assert merge_items([first, second]) == [first, second]
            assert merge_items([second, first]) == [second, first]
            assert merge_items([first, second, first]) == [first, second, first]

    def test_merge_union_of_same_length_tuples(self):
        # type: () -> None
        assert merge_items([TupleType([CT('str')]),
                            TupleType([CT('int')])]) == [TupleType([UnionType([CT('str'),
                                                                               CT('int')])])]
        assert merge_items([TupleType([CT('str')]),
                            TupleType([CT('Text')])]) == [TupleType([CT('Text')])]

    def test_merge_tuples_with_different_lengths(self):
        # type: () -> None
        assert merge_items([
            TupleType([CT('str')]),
            TupleType([CT('str'), CT('str')])]) == [CT('Tuple', [CT('str')])]
        assert merge_items([
            TupleType([]),
            TupleType([CT('str')]),
            TupleType([CT('str'), CT('str')])]) == [CT('Tuple', [CT('str')])]
        # Don't merge if types aren't identical
        assert merge_items([
            TupleType([CT('str')]),
            TupleType([CT('str'), CT('int')])]) == [TupleType([CT('str')]),
                                                    TupleType([CT('str'), CT('int')])]

    def test_merge_union_containing_no_return(self):
        # type: () -> None
        assert merge_items([CT('int'), NoReturnType()]) == [CT('int')]
        assert merge_items([NoReturnType(), CT('int')]) == [CT('int')]


class TestFlattenTypes(unittest.TestCase):
    def test_nested_tuples(self):
        # type: () -> None
        assert flatten_types([UnionType([UnionType([CT('int'), CT('str')]), CT('X')])]) == [
            CT('int'), CT('str'), CT('X')]



# License for Sphinx
# ==================
#
# Copyright (c) 2007-2017 by the Sphinx team (see AUTHORS file).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, print_function

import collections
import re
from typing import (Any, Callable, Deque, Generic, Iterable, Iterator, List,
                    Optional, Tuple, TypeVar, Union, overload)

from six import string_types

from typeright.docs.parsers import Arg

T = TypeVar('T')

SENTINEL = (None, None)

_directive_regex = re.compile(r'\.\. \S+::')
_google_section_regex = re.compile(r'^(\s|\w)+:\s*$')
_google_typed_arg_regex = re.compile(r'\s*(.+?)\s*\(\s*(.+?)\s*\)')
_numpy_section_regex = re.compile(r'^[=\-`:\'"~^_*+#<>]{2,}\s*$')
_xref_regex = re.compile(r'(:\w+:\S+:`.+?`|:\S+:`.+?`|`.+?`)')
_bullet_list_regex = re.compile(r'^(\*|\+|\-)(\s+\S|\s*$)')
_enumerated_list_regex = re.compile(
    r'^(?P<paren>\()?'
    r'(\d+|#|[ivxlcdm]+|[IVXLCDM]+|[a-zA-Z])'
    r'(?(paren)\)|\.)(\s+\S|\s*$)')


class peek_iter(Generic[T]):
    """An iterator object that supports peeking ahead.

    Parameters
    ----------
    obj : Iterable[T]
        `o` is interpreted very differently depending on the presence of
        `sentinel`.

        If `sentinel` is not given, then `o` must be a collection object
        which supports either the iteration protocol or the sequence protocol.

        If `sentinel` is given, then `o` must be a callable object.

    sentinel : Optional[Any]
        If given, the iterator will call `o` with no arguments for each
        call to its `next` method; if the value returned is equal to
        `sentinel`, :exc:`StopIteration` will be raised, otherwise the
        value will be returned.

    See Also
    --------
    `peek_iter` can operate as a drop in replacement for the built-in
    `iter <https://docs.python.org/2/library/functions.html#iter>`_ function.

    Attributes
    ----------
    sentinel
        The value used to indicate the iterator is exhausted. If `sentinel`
        was not given when the `peek_iter` was instantiated, then it will
        be set to a new object instance: ``object()``.

    """

    def __init__(self, obj, sentitnel=None):
        # type: (Iterable[T], Any) -> None
        self._iterable = iter(obj)
        self._cache = collections.deque()  # type: Deque[T]
        if sentitnel is not None:
            self.sentinel = sentitnel
        else:
            self.sentinel = object()

    def __iter__(self):
        # type: () -> Iterator[T]
        return self

    @overload
    def __next__(self):
        # type: () -> T
        pass

    @overload
    def __next__(self, n):
        # type: (int) -> List[T]
        pass

    def __next__(self, n=None):
        # notype
        # note: prevent 2to3 to transform self.next() in next(self) which
        # causes an infinite loop !
        return getattr(self, 'next')(n)

    def _fillcache(self, n):
        """Cache `n` items. If `n` is 0 or None, then 1 item is cached."""
        if not n:
            n = 1
        try:
            while len(self._cache) < n:
                self._cache.append(next(self._iterable))
        except StopIteration:
            while len(self._cache) < n:
                self._cache.append(self.sentinel)

    def has_next(self):
        # type: () -> bool
        """Determine if iterator is exhausted.

        Returns
        -------
        bool
            True if iterator has more items, False otherwise.

        Note
        ----
        Will never raise :exc:`StopIteration`.
        """
        return self.peek() != self.sentinel

    @overload
    def next(self):
        # type: () -> T
        pass

    @overload
    def next(self, n):
        # type: (int) -> List[T]
        pass

    def next(self, n=None):
        # type: (Optional[int]) -> Union[T, List[T]]
        # notype
        """Get the next item or `n` items of the iterator.

        Parameters
        ----------
        n : Optional[int]
            The number of items to retrieve. Defaults to None.

        Returns
        -------
        Union[T, List[T]]
            The next item or `n` items of the iterator. If `n` is None, the
            item itself is returned. If `n` is an int, the items will be
            returned in a list. If `n` is 0, an empty list is returned.

        Raises
        ------
        StopIteration
            Raised if the iterator is exhausted, even if `n` is 0.
        """
        self._fillcache(n)
        if not n:
            if self._cache[0] == self.sentinel:
                raise StopIteration
            if n is None:
                result = self._cache.popleft()  # type: Union[T, List[T]]
            else:
                result = []
        else:
            if self._cache[n - 1] == self.sentinel:
                raise StopIteration
            result = [self._cache.popleft() for i in range(n)]
        return result

    @overload
    def peek(self):
        # type: () -> T
        pass

    @overload
    def peek(self, n):
        # type: (int) -> List[T]
        pass

    def peek(self, n=None):
        # type: (Optional[int]) -> Union[T, List[T]]
        # notype
        """Preview the next item or `n` items of the iterator.

        The iterator is not advanced when peek is called.

        Parameters
        ----------
        n : Optional[int]

        Returns
        -------
        Union[T, List[T]]
            The next item or `n` items of the iterator. If `n` is None, the
            item itself is returned. If `n` is an int, the items will be
            returned in a list. If `n` is 0, an empty list is returned.

            If the iterator is exhausted, `peek_iter.sentinel` is returned,
            or placed as the last item in the returned list.

        Note
        ----
        Will never raise :exc:`StopIteration`.
        """
        self._fillcache(n)
        if n is None:
            result = self._cache[0]  # type: Union[T, List[T]]
        else:
            result = [self._cache[i] for i in range(n)]
        return result


class modify_iter(peek_iter[T]):
    """An iterator object that supports modifying items as they are returned.

    Parameters
    ----------
    obj : Iterable[Any]
        `o` is interpreted very differently depending on the presence of
        `sentinel`.

        If `sentinel` is not given, then `o` must be a collection object
        which supports either the iteration protocol or the sequence protocol.

        If `sentinel` is given, then `o` must be a callable object.

    sentinel : Optional[Any]
        If given, the iterator will call `o` with no arguments for each
        call to its `next` method; if the value returned is equal to
        `sentinel`, :exc:`StopIteration` will be raised, otherwise the
        value will be returned.

    modifier : Callable[[Any], T]
        The function that will be used to modify each item returned by the
        iterator. `modifier` should take a single argument and return a
        single value. Defaults to ``lambda x: x``.

        If `sentinel` is not given, `modifier` must be passed as a keyword
        argument.

    Attributes
    ----------
    modifier : Callable
        `modifier` is called with each item in `o` as it is iterated. The
        return value of `modifier` is returned in lieu of the item.

        Values returned by `peek` as well as `next` are affected by
        `modifier`. However, `modify_iter.sentinel` is never passed through
        `modifier`; it will always be returned from `peek` unmodified.

    Example
    -------
    >>> a = ["     A list    ",
    ...      "   of strings  ",
    ...      "      with     ",
    ...      "      extra    ",
    ...      "   whitespace. "]
    >>> modifier = lambda s: s.strip().replace('with', 'without')
    >>> for s in modify_iter(a, modifier=modifier):
    ...   print('"%s"' % s)
    "A list"
    "of strings"
    "without"
    "extra"
    "whitespace."

    """

    def __init__(self, obj, modifier, sentinel=None):
        # type: (Iterable[Any], Callable[[Any], T], Optional[Any]) -> None
        self.modifier = modifier
        if not callable(self.modifier):
            raise TypeError('modify_iter(o, modifier): '
                            'modifier must be callable')
        super(modify_iter, self).__init__(obj, sentinel)

    def _fillcache(self, n):
        """Cache `n` modified items. If `n` is 0 or None, 1 item is cached.

        Each item returned by the iterator is passed through the
        `modify_iter.modified` function before being cached.

        """
        if not n:
            n = 1
        try:
            while len(self._cache) < n:
                self._cache.append(self.modifier(next(self._iterable)))
        except StopIteration:
            while len(self._cache) < n:
                self._cache.append(self.sentinel)


class GoogleDocstring(object):

    _directive_sections = []  # type: List[str]

    def __init__(self, docstring):
        # type: (Union[str, List[str]]) -> None
        if isinstance(docstring, string_types):
            docstring = docstring.splitlines()
        self._lineno = 1
        self._lines = docstring

        def next_line(line):
            # type: (str) -> Tuple[str, int]
            self._lineno += 1
            return line.rstrip(), self._lineno

        self._line_iter = modify_iter(docstring, next_line, SENTINEL)
        self._is_in_section = False
        self._section_indent = 0

        self._params = None  # type: Optional[List[Tuple[str, Arg]]]
        self._returns = None  # type: Optional[List[Arg]]
        self._yields = None  # type: Optional[List[Arg]]

        self._sections = {
            'args': self._parse_parameters_section,
            'arguments': self._parse_parameters_section,
            'attributes': self._skip_section,
            'example': self._skip_section,
            'examples': self._skip_section,
            'keyword args': self._skip_section,
            'keyword arguments': self._skip_section,
            'methods': self._skip_section,
            'note': self._skip_section,
            'notes': self._skip_section,
            'other parameters': self._skip_section,
            'parameters': self._parse_parameters_section,
            'return': self._parse_returns_section,
            'returns': self._parse_returns_section,
            'raises': self._skip_section,
            'references': self._skip_section,
            'see also': self._skip_section,
            'todo': self._skip_section,
            'warning': self._skip_section,
            'warnings': self._skip_section,
            'warns': self._skip_section,
            'yield': self._parse_yields_section,
            'yields': self._parse_yields_section,
        }

    def _consume_indented_block(self, indent=1):
        # type: (int) -> List[Tuple[str, int]]
        lines = []  # type: List[Tuple[str, int]]
        line = self._line_iter.peek()
        while(not self._is_section_break() and
              (not line[0] or self._is_indented(line[0], indent))):
            lines.append(next(self._line_iter))
            line = self._line_iter.peek()
        return lines

    def _consume_contiguous(self):
        # type: () -> List[Tuple[str, int]]
        lines = []  # type: List[Tuple[str, int]]
        while (self._line_iter.has_next() and
               self._line_iter.peek() and
               not self._is_section_header()):
            lines.append(next(self._line_iter))
        return lines

    def _consume_empty(self):
        # type: () -> List[Tuple[str, int]]
        lines = []  # type: List[Tuple[str, int]]
        line = self._line_iter.peek()
        while self._line_iter.has_next() and not line[0]:
            lines.append(next(self._line_iter))
            line = self._line_iter.peek()
        return lines

    def _consume_field(self, parse_type=True, prefer_type=False):
        # type: (bool, bool) -> Tuple[str, Optional[str], int]
        line, lineno = next(self._line_iter)

        before, colon, after = self._partition_field_on_colon(line, lineno)
        _name = before
        _type = None

        if parse_type:
            match = _google_typed_arg_regex.match(before)
            if match:
                _name = match.group(1)
                _type = match.group(2)

        _name = self._escape_args_and_kwargs(_name)

        if prefer_type and not _type:
            _type, _name = _name, _type
        indent = self._get_indent(line) + 1
        self._consume_indented_block(indent)
        return _name, _type, lineno

    def _consume_fields(self, parse_type=True, prefer_type=False):
        # type: (bool, bool) -> List[Tuple[str, Arg]]
        self._consume_empty()
        fields = []
        while not self._is_section_break():
            _name, _type, lineno = self._consume_field(parse_type, prefer_type)
            if _name or _type:
                fields.append((_name, Arg(_type, lineno)))
        return fields

    def _consume_returns_section(self):
        # type: () -> List[Arg]
        lines = self._dedent(self._consume_to_next_section())
        if lines:
            line, lineno = lines[0]
            before, colon, after = self._partition_field_on_colon(line, lineno)
            _type = None

            if colon:
                match = _google_typed_arg_regex.match(before)
                if match:
                    _type = match.group(2)
                else:
                    _type = before
            if _type:
                return [Arg(_type, lineno)]

        return []

    def _consume_section_header(self):
        # type: () -> str
        section, _ = next(self._line_iter)
        stripped_section = section.strip(':')
        if stripped_section.lower() in self._sections:
            section = stripped_section
        return section

    def _consume_to_end(self):
        # type: () -> List[Tuple[str, int]]
        lines = []  # type: List[Tuple[str, int]]
        while self._line_iter.has_next():
            lines.append(next(self._line_iter))
        return lines

    def _consume_to_next_section(self):
        # type: () -> List[Tuple[str, int]]
        lines = []  # type: List[Tuple[str, int]]
        self._consume_empty()
        while not self._is_section_break():
            lines.append(next(self._line_iter))
        return lines + self._consume_empty()

    def _dedent(self, lines, full=False):
        # type: (List[Tuple[str, int]], bool) -> List[Tuple[str, int]]
        if full:
            return [(line.lstrip(), lineno) for line, lineno in lines]
        else:
            min_indent = self._get_min_indent(lines)
            return [(line[min_indent:], lineno) for line, lineno in lines]

    def _escape_args_and_kwargs(self, name):
        # type: (str) -> str
        if name[:2] == '**':
            return r'\*\*' + name[2:]
        elif name[:1] == '*':
            return r'\*' + name[1:]
        else:
            return name

    def _get_current_indent(self, peek_ahead=0):
        # type: (int) -> int
        line = self._line_iter.peek(peek_ahead + 1)[peek_ahead]
        while line != self._line_iter.sentinel:
            if line[0]:
                return self._get_indent(line[0])
            peek_ahead += 1
            line = self._line_iter.peek(peek_ahead + 1)[peek_ahead]
        return 0

    def _get_indent(self, line):
        # type: (str) -> int
        for i, s in enumerate(line):
            if not s.isspace():
                return i
        return len(line)

    def _get_min_indent(self, lines):
        # type: (List[Tuple[str, int]]) -> int
        min_indent = None
        for line, _ in lines:
            if line:
                indent = self._get_indent(line)
                if min_indent is None:
                    min_indent = indent
                elif indent < min_indent:
                    min_indent = indent
        return min_indent or 0

    def _is_indented(self, line, indent=1):
        # type: (str, int) -> bool
        for i, s in enumerate(line):
            if i >= indent:
                return True
            elif not s.isspace():
                return False
        return False

    def _is_section_header(self):
        # type: () -> bool
        section = self._line_iter.peek()[0].lower()
        match = _google_section_regex.match(section)
        if match and section.strip(':') in self._sections:
            header_indent = self._get_indent(section)
            section_indent = self._get_current_indent(peek_ahead=1)
            return section_indent > header_indent
        elif self._directive_sections:
            if _directive_regex.match(section):
                for directive_section in self._directive_sections:
                    if section.startswith(directive_section):
                        return True
        return False

    def _is_section_break(self):
        # type: () -> bool
        line, _ = self._line_iter.peek()
        return bool(
            not self._line_iter.has_next() or
            self._is_section_header() or
            (self._is_in_section and
                line and
                not self._is_indented(line, self._section_indent)))

    def parse(self):
        # type: () -> Tuple[Optional[List[Tuple[str, Arg]]], Optional[List[Arg]], Optional[List[Arg]]]
        self._consume_empty()

        while self._line_iter.has_next():
            if self._is_section_header():
                try:
                    section = self._consume_section_header()
                    self._is_in_section = True
                    self._section_indent = self._get_current_indent()
                    if _directive_regex.match(section):
                        self._consume_to_next_section()
                    else:
                        self._sections[section.lower()]()
                finally:
                    self._is_in_section = False
                    self._section_indent = 0
            else:
                # FIXME: behavior varied based on parsed lines, which we removed
                # if not self._parsed_lines:
                #     self._consume_contiguous() + self._consume_empty()
                # else:
                self._consume_to_next_section()

        return self._params, self._returns, self._yields

    def _skip_section(self):
        self._consume_to_next_section()

    def _parse_parameters_section(self):
        self._params = self._consume_fields()

    def _parse_returns_section(self):
        self._returns = self._consume_returns_section()

    def _parse_yields_section(self):
        self._yields = self._consume_returns_section()

    def _partition_field_on_colon(self, line, lineno):
        # type: (str, int) -> Tuple[str, str, str]
        before_colon = []
        after_colon = []
        colon = ''
        found_colon = False
        for i, source in enumerate(_xref_regex.split(line)):
            if found_colon:
                after_colon.append(source)
            else:
                if (i % 2) == 0 and ":" in source:
                    found_colon = True
                    before, colon, after = source.partition(":")
                    before_colon.append(before)
                    after_colon.append(after)
                else:
                    before_colon.append(source)

        return ("".join(before_colon).strip(),
                colon,
                "".join(after_colon).strip())


class NumpyDocstring(GoogleDocstring):

    _directive_sections = ['.. index::']

    def _consume_field(self, parse_type=True, prefer_type=False):
        line, lineno = next(self._line_iter)
        if parse_type:
            _name, _, _type = self._partition_field_on_colon(line, lineno)
        else:
            _name = line
            _type = None

        _name, _type = _name.strip(), _type.strip()
        _name = self._escape_args_and_kwargs(_name)

        if prefer_type and not _type:
            _type, _name = _name, _type
        indent = self._get_indent(line) + 1
        self._consume_indented_block(indent)
        return _name, _type, lineno

    def _consume_returns_section(self):
        # type: () -> List[Arg]
        fields = self._consume_fields(prefer_type=True)
        return [x[1] for x in fields]

    def _consume_section_header(self):
        section, _ = next(self._line_iter)
        if not _directive_regex.match(section):
            # Consume the header underline
            next(self._line_iter)
        return section

    def _is_section_break(self):
        line1, line2 = self._line_iter.peek(2)
        return (not self._line_iter.has_next() or
                self._is_section_header() or
                ['', ''] == [line1[0], line2[0]] or
                (self._is_in_section and
                    line1[0] and
                    not self._is_indented(line1[0], self._section_indent)))

    def _is_section_header(self):
        section, underline = self._line_iter.peek(2)
        section = section[0].lower()
        underline = underline[0]
        if section in self._sections and isinstance(underline, string_types):
            return bool(_numpy_section_regex.match(underline))
        elif self._directive_sections:
            if _directive_regex.match(section):
                for directive_section in self._directive_sections:
                    if section.startswith(directive_section):
                        return True
        return False

    _name_rgx = re.compile(r"^\s*(:(?P<role>\w+):`(?P<name>[a-zA-Z0-9_.-]+)`|"
                           r" (?P<name2>[a-zA-Z0-9_.-]+))\s*", re.X)
